
	/* 
	 * These functions deal with parsing a command string
	 * and then calling the appropriate function.
	 */

#include <stdio.h>
#include <string.h>
#include <unistd.h>      /* for usleep() */
#include <math.h>
#include "tait.h"
#include "radec.h"


/* stuff used to create a hash table of commands and their functions */
#define HASHSIZE 100
typedef int (*PFI)();
typedef struct s_hash {
	PFI func;
	char name[LINELEN];
	struct s_hash *next;
} HASH;
static HASH *hashtab[HASHSIZE];


	/* functions private to this source code file */
static PFI lookup(char *cmd);
static void install(char *name, PFI funcp);
static int calc_hash(char *name);
static void strip_newline(char *str);

	/* these are functions to handle commands from the user */
static int cmd_quit(char *cmdline);
static int cmd_help(char *cmdline);
static int cmd_nop(char *cmdline);
static int cmd_echo(char *cmdline);
static int cmd_ra(char *cmdline);
static int cmd_motor(char *cmdline);
static int cmd_clamp(char *cmdline);
static int cmd_dec(char *cmdline);
static int cmd_shutter(char *cmdline);
static int cmd_focus(char *cmdline);
static int cmd_exp(char *cmdline);
static int cmd_show(char *cmdline);
static int cmd_set(char *cmdline);
static int cmd_saveimg(char *cmdline);
static int cmd_radec(char *cmdline);

	/* low-level functions for the Dec drive */
static int dec_move_raw(int steps);
static int dec_move_to_limit(void);
static int dec_move_rel(int steps);
static int dec_move_abs(int steps);

	/* low-level functions for the focus motors */
static int foc_to_limit(MOTOR_ID id);
static int foc_move_rel(MOTOR_ID id, int steps);
static int foc_move_abs(MOTOR_ID id, int steps);

	/* low-level functions to move to given (RA, Dec) */
static int goto_radec(double ra, double dec);



	/**************************************************************
	 * PROCEDURE: read_next_command
	 *
	 * DESCRIPTION: Read a line's worth of information, 
	 *              or from an input file (with the given name),
	 *              or stdin (if the 'input_file' arg is empty).
	 *
	 *              Place the line we read into the "command_line"
	 *              argument (which must have sufficient space
	 *              allocated).
	 *
	 * RETURNS:
	 *    0           if we did read a command
	 *    1           if we have reached end of input
	 */

int
read_next_command
	(
	char *input_file,      /* I: either the file name, or */
	                       /*       '\0' to signal stdin */
	char *command_line     /* O: place the information we read here */
	)
{
	if (input_file[0] == '\0') {
		if (fgets(command_line, LINELEN, stdin) == NULL) {
			return(1);
		}
	}
	else {
		if (get_next_line(input_file, command_line) != 0) {
			shError("read_next_command: get_next_line fails");
			return(1);
		}
	}
	return(0);
}

	/**************************************************************
	 * PROCEDURE: parse_and_execute
	 *
	 * DESCRIPTION: Given a command-line string, like
	 *                     exp dark 15
	 *              or
	 *                     quit
	 *
	 *              this function splits the line into words,
	 *              uses the first word to look up the command,
	 *              then calls the appropriate function
	 *              with appropriate arguments.
	 *
	 *              If the command is "quit" or "exit",
	 *              then this function returns 1 to indicate
	 *              "the program should halt."
	 *              Otherwise, it returns 0, even if the command
	 *              wasn't recognized.
	 *
	 * RETURNS:
	 *      0          if the command isn't to quit
	 *      1          if the command means we should quit
	 */

int
parse_and_execute
	(
	char *cmdline      /* I: an ASCII string, command plus args */
	)
{
	char cmdword[LINELEN+1];
	int retval;
	PFI func;

	shAssert(cmdline != NULL);


	/* 
	 * echo the entire command line to the log;
	 * we need to strip the newline from it first, though.
	 */
	cmdline[LINELEN] = '\0';
	strip_newline(cmdline);
	shLog("%s", cmdline);

	/* 
	 * get the first word of the command line and look it up.
	 * if we recognize it, call the function corresponding to it.
	 * if it isn't recognized, print an error message and return.
	 */
	if (sscanf(cmdline, "%s", cmdword) != 1) {
		/* must have been a blank line.  Just skip it silently */
		return(PARSE_OK);
	}
	if ((func = lookup(cmdword)) != NULL) {
		/* execute the function */
		shLog("  execute ..%s..", cmdword);
		retval = (*func)(cmdline);
		return(retval);
	} else {
		/* unrecognized command */
		shError("  unrecognized command ..%s..", cmdword);
		return(PARSE_ERROR);
	}


	return(PARSE_OK);
}



	/**********************************************************************
	 * PROCEDURE: init_hash
	 *
	 * DESCRIPTION: Initialize the hash table, first filling it with empty
	 *              entries, then installing all commands.
	 *
	 * RETURNS:
	 *     nothing
	 */

void
init_hash
	(
	void
	)
{
	int i;

	/* first, fill with empty entries */
	for (i = 0; i < HASHSIZE; i++) {
		hashtab[i] = NULL;
	}

	/* now, install all known commands */
	install("quit", cmd_quit);
	install("end", cmd_quit);
	install("exit", cmd_quit);
	install("help", cmd_help);
	install("nop", cmd_nop);
	install("onp", cmd_nop);
	install("pon", cmd_nop);
	install("echo", cmd_echo);
	install("#", cmd_echo);
	install("ra", cmd_ra);
	install("motor", cmd_motor);
	install("clamp", cmd_clamp);
	install("dec", cmd_dec);
	install("shutter", cmd_shutter);
	install("focus", cmd_focus);
	install("exp", cmd_exp);
	install("show", cmd_show);
	install("set", cmd_set);
	install("saveimg", cmd_saveimg);
	install("radec", cmd_radec);

}
	

	/**********************************************************************
	 * PROCEDURE: lookup
	 *
	 * DESCRIPTION: Given a string with the name of a command, like
	 *              "exp" or "ra", look it up in the hash table.
	 *              If we find a match, return a pointer to the 
	 *              appropriate function.  If not, return NULL.
	 *
	 * RETURNS:
	 *     pointer-to-function          if command is found
	 *     NULL                         if not
	 */

static PFI
lookup
	(
	char *cmd                 /* I: name of command to search for */
	)
{
	int index;
	HASH *hashp;

	index = calc_hash(cmd);
	hashp = hashtab[index];
	while (hashp != NULL) {
		if (strncmp(hashp->name, cmd, LINELEN) == 0) {
#ifdef DEBUG
			printf("  found match for cmd ..%s.. is ..%s..\n", 
						cmd, hashp->name);
#endif
			return(hashp->func);
		}
#ifdef DEBUG
		printf("  no match with entry ..%s.., so go to next pointer\n", 
				hashp->name);
#endif
		hashp = hashp->next;
	}

	/* if we get here, the command isn't in the table. */
#ifdef DEBUG
	printf("  nope, no match for ..%s.. \n", cmd);
#endif
	return(NULL);

}


	/***********************************************************************
	 * PROCEDURE: install
	 *
	 * DESCRIPTION: Place an entry into the hash table, associating
	 *              a command name with a pointer to its function.
	 *              If there is already a command in the table at the
	 *              position of this name's hash, place the entry
	 *              for this function at the end of a linked list
	 *              starting at that position.
	 *
	 * RETURNS:
	 *     nothing
	 */

static void
install
	(
	char *name,            /* I: name of the command */
	PFI funcp              /* I: pointer to the appropriate function */
	)
{
	int index;
	HASH *hashp, *new;

	index = calc_hash(name);
	hashp = hashtab[index];
	while ((hashp != NULL) && (hashp->func != NULL)) {
		/* 
		 * if we have already installed a function for this name,
		 * print an error message and return without installing
		 * a second entry.
		 */
		if (strncmp(name, hashp->name, LINELEN) == 0) {
			shError("install: ignoring duplicate entry %s", name);
			return;
		}

		if (hashp->next == NULL) {
			break;
		} else {
			hashp = hashp->next;
		}
	}

	/* 
	 * if we get here, 'hashp' points to the current end of the
	 * linked list of hash structures.  We need to add an entry
	 * for this command 
	 */
	new = (HASH *) shMalloc(sizeof(HASH));
	strncpy(new->name, name, LINELEN);
	new->name[LINELEN] = '\0';
	new->func = funcp;
	new->next = NULL;

	if (hashp == NULL) {
		hashtab[index] = new;
	} 
	else {
		hashp->next = new;
	}

}



	/**********************************************************************
	 * PROCEDURE: calc_hash
	 *
	 * DESCRIPTION: given a string with a single command, like "dec"
	 *              or "quit", calculate a hash value.
	 *
	 * RETURNS:
	 *    the hash value
	 */

static int
calc_hash
	(
	char *name           /* I: string with word for us to hash */
	)
{
	int i, index;

	/* hash the command into an index */
	index = 0;
	for (i = 0; i < strlen(name); i++) {
		index += name[i];
	}
	index = index % HASHSIZE;
#ifdef DEBUG
	printf("calc_hash: name ..%s.. has index %d \n", name, index);
#endif

	return(index);


}



	/**********************************************************************
	 * PROCEDURE: strip_newline
	 *
	 * DESCRIPTION: Given a string, replace any newline chars in it
	 *              with blank spaces.
	 *
	 * RETURNS: 
	 *    nothing
	 */

static void
strip_newline
	(
	char *str              /* I/O: replace newlines in this string */
	)
{
	int i;

	for (i = 0; str[i] != '\0'; i++) {
		if (str[i] == '\n') {
			str[i] = ' ';
		}
	}
}




	/*********************************************************************
	 * PROCEDURE: cmd_quit
	 *
	 * DESCRIPTION: Handle the "quit" command.
	 *
	 *              Usage:     quit
	 *
	 *
	 *              Return the PARSE_QUIT value, signalling that
	 *              the main program should terminate.
	 *
	 * RETURNS:
	 *     PARSE_QUIT
	 */

static int
cmd_quit
	(
	char *cmdline              /* I: the entire command line, */
	                           /*    including the command word */
	)
{

	return(PARSE_QUIT);
}



	/*********************************************************************
	 * PROCEDURE: cmd_nop
	 *
	 * DESCRIPTION: Handle the "nop" command.
	 *
	 *              Usage:     nop ...
	 *
	 *              Do nothing ("nop" = "no operation").
	 *              
	 *              (This is a good place to put snippets
	 *              of code for test execution.)
	 *
	 * RETURNS:
	 *     PARSE_OK
	 */

static int
cmd_nop
	(
	char *cmdline              /* I: the entire command line, */
	                           /*    including the command word */
	)
{
	struct tm tms;
	double jd;
	double lst;
	int lst_h, lst_m, lst_s;

	get_now(&tms);
	printf("  UT time is %02d:%02d:%02d \n", tms.tm_hour, tms.tm_min, 
					tms.tm_sec);
	get_now_jd(&jd);
	printf("  JD is %12.6f \n", jd);
	get_lst(jd, &lst);
	lst_h = floor(lst);
	lst_m = floor((lst - lst_h)*60.0);
	lst_s = floor((lst - (lst_h + lst_m/60.0))*3600.0);

	printf("  LST is %9.5f = %02d:%02d:%02d \n", lst, lst_h, lst_m, lst_s);

	shLog("OK nop");

	return(PARSE_OK);
}


	/*********************************************************************
	 * PROCEDURE: cmd_echo
	 *
	 * DESCRIPTION: Handle the "echo" command.
	 *
	 *              Usage:     echo  ...
	 *
	 *              Simply echo the arguments and return.
	 *              
	 * RETURNS:
	 *     PARSE_OK
	 */

static int
cmd_echo
	(
	char *cmdline              /* I: the entire command line, */
	                           /*    including the command word */
	)
{
	shLog("  %s", cmdline);
	shLog("OK echo");

	return(PARSE_OK);
}




	/*********************************************************************
	 * PROCEDURE: cmd_help
	 *
	 * DESCRIPTION: Handle the "help" command.
	 *
	 *              Usage:     help
	 *
	 *              Print a list of all the commands we know,
	 *              by looking through the hash table.
	 *
	 * RETURNS:
	 *     PARSE_OK       if proper arguments given
	 */

static int
cmd_help
	(
	char *cmdline              /* I: the entire command line, */
	                           /*    including the command word */
	)
{
	int i, j, k;
	char reply[LINELEN+1];
	HASH *hashp;

	strcpy(reply, "OK help:  ");

	for (i = 0; i < HASHSIZE; i++) {
		hashp = hashtab[i];
		while (hashp != NULL) {

			/* 
			 * if the list of commands would overflow 'reply',
			 * then print current 'reply' and start it over.
			 */
			if (strlen(reply) + strlen(hashp->name) + 1 > LINELEN) {
				shLog("%s", reply);
				strcpy(reply, "  ");
			}

			/* append the current command name to 'reply' string */
			for (j = 0; reply[j] != '\0'; j++) {
				;
			}
			for (k = 0; hashp->name[k] != '\0'; k++) {
				reply[j + k] = hashp->name[k];
			}
			reply[j + k++] = ' ';
			reply[j + k++] = '\0';

			hashp = hashp->next;
		}
	}
	shLog("%s", reply);

	return(PARSE_OK);
}


	/*********************************************************************
	 * PROCEDURE: cmd_ra
	 *
	 * DESCRIPTION: Handle the "ra" command.
	 *
	 *              Usage:     ra  [on|off|ff|rev|atlimit] 
	 *
	 *
	 *              Control the RA motor: 
	 *                  off           turns it off
	 *                  on            turns it on at sidereal rate:
	 *                                     moves West to follow stars,
	 *                                     ordinary tracking mode
	 *                  ff            moves it quickly West
	 *                  rev           moves it quickly East
	 *                  atlimit       checks to see if RA drive at limit
	 *                                     (pointing maximum East)
	 *
	 * This routine turns on the RA drive motor, so that it makes the
	 * telescope track at sidereal speed to the west.
	 *
	 * In the control lines below, variable f$ controls speed and direction
	 * of the motor.  The variable has 8 bits, grouped like so:
	 *
	 *                     xx yyy ww z
	 * where
	 *         highest two bits     xx     not used
	 *         three middle bits    yyy    indicate speed of motor
	 *                                        110 = sidereal, 111 = faster
	 *         two middle bits      ww     not sure ("countdown bits?")
	 *         lowest bit           z      gives direction of motor:
	 *                                        0 = forward, 1 = reverse
	 *
	 * The variable g$ controls the current flowing through the RA Motor
	 *                "0"   turns on current
	 *                "8"   turns off current
	 *
	 * Following are some cryptic comments found in the original code,
	 * which I believe are different representations of the value
	 * of this variable f$ which controls RA motor speed and direction.
	 *
	 * 0,1     0.1422'2,3     0.071149
	 * 4,5     0.035574
	 * 6,7     0.017787
	 * 8,9     0.008893
	 * :,;     0.004446
	 * <,=     0.002223
	 * >,?     External
	 *
	 *
	 * RETURNS:
	 *     PARSE_OK       if proper arguments given
	 *     PARSE_ERROR    if improper arguments
	 */

static int
cmd_ra
	(
	char *cmdline              /* I: the entire command line, */
	                           /*    including the command word */
	)
{
	char cmd[LINELEN+1], arg[LINELEN+1];
	char reply[MSG_LEN+1];
	int retval;

	if (sscanf(cmdline, "%s %s", cmd, arg) != 2) {
		shError("cmd_ra: wrong number of args");
		return(PARSE_ERROR);
	}
	arg[LINELEN] = '\0';

	if (strcmp(arg, "help") == 0) {
		shLog("OK  ra off, on, ff, rev, atlimit");
	}
	else if (strcmp(arg, "off") == 0) {
		/* 
		 * loads 111 in count down selection 
		 *
		 * the "8" in 7'th char turns off motor current 
		 */
		send_message("581;;?83", reply, "turn RA off");
		shLog("OK ra off");
	}
	else if (strcmp(arg, "on") == 0) {
		/* 
		 * the "0" in 6'th char was 4 for -014 motorcount 
		 *      down factors of two and direction  
		 *
		 * the "0" in 7'th char turns on the motor current 
		 */
		send_message("581;;003", reply, "turn RA on");
		shLog("OK ra on");
	}
	else if (strcmp(arg, "ff") == 0) {
		/* 
		 * Note that we must send two messages to the STAMP.
		 *
		 * Comment on first call:
		 *    if "6" placed in 6'th char, motor would 
		 *    move at 8x normal speed.
		 *
		 * Comment on second call:
		 *    the ":" in 6'th char is faster but not quite
		 *    enough torque
		 */
		send_message("581;;803", reply, "turn RA ff 1");
		send_message("581;;:03", reply, "turn RA ff 2");
		shLog("OK ra ff");
	}
	else if (strcmp(arg, "rev") == 0) {
		/* 
		 * safety check: if RA at home position, refuse
		 * to move any further East.
		 */
		if (ra_is_at_limit() == 1) {
			shLog("OK RA already at limit, no rev allowed");
		}
		else {
			/* 
		 	* Note that we must send two messages to the STAMP.
		 	*
		 	* Comment on first call:
		 	*    if "7" placed in 6'th char, motor would 
		 	*    move at 8x normal speed.
		 	*
		 	* Comment on second call:
		 	*    the ":" in 6'th char is faster but not quite
		 	*    enough torque
		 	*/
			send_message("581;;903", reply, "turn RA rev 1");
			send_message("581;;;03", reply, "turn RA rev 2");
			shLog("OK ra rev");
		}
	}
	else if (strcmp(arg, "atlimit") == 0) {
		retval = ra_is_at_limit();
		if (retval == 1) {
			shLog("OK RA limit yes");
		} 
		else {
			shLog("OK RA limit no");
		}
	}
	else if ((strcmp(arg, "tolimit") == 0) ||
	         (strcmp(arg, "home") == 0)) {
		if (ra_is_at_limit() == 0) {
			/* move the RA motor quickly in reverse */
			send_message("581;;903", reply, "turn RA rev 1");
			send_message("581;;;03", reply, "turn RA rev 2");
			while (ra_is_at_limit() == 0) {
#ifdef DEBUG
				printf("          RA still not at limit\n");
#endif
				;
			}
			/* now stop the RA motor */
			send_message("581;;?83", reply, "turn RA off");
			/* 
			 * remove backlash by moving forward (west)
			 * at ordinary speed for 1 second.
			 */
			send_message("581;;003", reply, "turn RA on");
			sleep(1);
			send_message("581;;?83", reply, "turn RA off");
			shLog("OK ra tolimit");
		}
	}
	else {
		shError("ra: unrecognized arg ..%s..", arg);
		return(PARSE_ERROR);
	}

	return(PARSE_OK);
}


	/*********************************************************************
	 * PROCEDURE: cmd_motor
	 *
	 * DESCRIPTION: Handle the "motor" command.  This permits the user
	 *              to send messages via a very low-level protocol,
	 *              which should probably be disabled in the final
	 *              product.  We enable it now for debugging and
	 *              development.
	 *
	 *              Usage:     motor  action id steps direction
	 *
	 *                where the 4 required args are all integers.
	 *
	 *                    action      can be MOTOR_LIMIT_OUT
	 *                                       MOTOR_IN_DIRECTION
	 *                                       MOTOR_OFF
	 *
	 *                    id          can be MOTOR_FOCUSV
	 *                                       MOTOR_FOCUSI
	 *                                       MOTOR_DECDRIVE
	 *                                       MOTOR_DECCLAMP
	 *
	 *                    steps       is the number of steps to move
	 *
	 *                    direction   can be MOTOR_TOWARD_LIMIT
	 *                                       MOTOR_FROM_LIMIT
	 *
	 * RETURNS:
	 *     PARSE_OK       if proper arguments given
	 *     PARSE_ERROR    if improper arguments
	 */

static int
cmd_motor
	(
	char *cmdline              /* I: the entire command line, */
	                           /*    including the command word */
	)
{
	char cmd[LINELEN+1];
	MOTOR_ACTION action;
	MOTOR_ID id;
	MOTOR_DIRECTION direction;
	int steps;
	int retval;

	if (sscanf(cmdline, "%s %d %d %d %d", cmd, 
				(int *)&action, 
				(int *)&id,
				(int *)&direction, 
				(int *)&steps) != 5) {
		shError("cmd_motor: wrong number of args");
		return(PARSE_ERROR);
	}

	else {
#ifdef DEBUG
		shLog("  about to call motor with args %d %d %d %d ",
				action, id, steps, direction);
#endif
		retval = motor(action, id, direction, steps);
#ifdef DEBUG
		shLog("  motor returns %d ", retval);
#endif

	}
	shLog("OK motor");
	return(PARSE_OK);
}


	/*********************************************************************
	 * PROCEDURE: cmd_clamp
	 *
	 * DESCRIPTION: Handle the "clamp" command.
	 *
	 *              Usage:     clamp    [on|off]
	 *
	 *
	 *              Control the Dec clamp motor: 
	 *                  off           loosens the clamp
	 *                  on            tightens the clamp
	 *
	 *              There is no limit switch for this motor,
	 *              but the convention is 
	 *                   MOTOR_TOWARD_LIMIT     tightens
	 *                   MOTOR_FROM_LIMIT       loosens
	 *
	 *              We move a greater distance when tightening
	 *              than loosening, to assure a solid connection.
	 *
	 *              The STAMP will return immediately after 
	 *              receiving the instructions which move the
	 *              Dec clamp -- without waiting for the motor
	 *              to finish (un)clamping the brake pad.  
	 *              We must wait until that occurs, for two reasons:
	 *                  a. don't want to try to move Dec axis if
	 *                     clamp is on
	 *                  b. STAMP may be confused if sent additional
	 *                     commands while motor is still moving
	 *
	 *              So, we wait explicitly after calling the 
	 *              motor routines, for a "safe" period.
	 *
	 * RETURNS:
	 *     PARSE_OK       if proper arguments given
	 *     PARSE_ERROR    if improper arguments
	 */

static int
cmd_clamp
	(
	char *cmdline              /* I: the entire command line, */
	                           /*    including the command word */
	)
{
	char cmd[LINELEN+1], arg[LINELEN+1];
	int retval;

	if (sscanf(cmdline, "%s %s", cmd, arg) != 2) {
		shError("clamp: wrong number of args");
		return(PARSE_ERROR);
	}
	arg[LINELEN] = '\0';

	if (strcmp(arg, "help") == 0) {
		shLog("OK clamp off, on ");
	}
	else if (strcmp(arg, "off") == 0) {
		/* 
		 * move the motor so the clamp loosens
		 */
		retval = motor(MOTOR_IN_DIRECTION, MOTOR_DECCLAMP, 
			MOTOR_FROM_LIMIT, 200);
		if (retval != 0) {
			shError("cmd_clamp: motor returns with error");
		}
	}
	else if (strcmp(arg, "on") == 0) {
		/* 
		 */
		retval = motor(MOTOR_IN_DIRECTION, MOTOR_DECCLAMP, 
			MOTOR_TOWARD_LIMIT, 250);
		if (retval != 0) {
			shError("cmd_clamp: motor returns with error");
		}
	}
	else {
		shError("clamp: unrecognized arg ..%s..", arg);
		return(PARSE_ERROR);
	}

	shLog("OK clamp");
	return(PARSE_OK);
}



	/*********************************************************************
	 * PROCEDURE: cmd_dec
	 *
	 * DESCRIPTION: Handle the "dec" command.
	 *
	 *              Usage:     dec    atlimit 
	 *                                tolimit   (or home)
	 *                                abs NN
	 *                                rel NN
	 *
	 *              where 
	 *                   atlimit        report the condition of the Dec 
	 *                                        axis limit switch
	 *
	 *                   tolimit        move Dec axis to limit switch
	 *                   home
	 *
	 *                   abs NN         move NN steps away from the 
	 *                                        limit switch position
	 *
	 *                   rel NN         move NN steps away from the 
	 *                                        current position
	 *
	 *               If NN is a positive value, move North
	 *                          negative             South
	 *
	 * RETURNS:
	 *     PARSE_OK       if proper arguments given
	 *     PARSE_ERROR    if improper arguments
	 */

static int
cmd_dec
	(
	char *cmdline              /* I: the entire command line, */
	                           /*    including the command word */
	)
{
	char cmd[LINELEN+1], arg[LINELEN+1];
	int narg = -1;
	int steps = -1;
	int retval;

	narg = sscanf(cmdline, "%s %s %d", cmd, arg, &steps);
	if ((narg < 2) || (narg > 3)) {
		shError("cmd_dec: wrong number of args");
		return(PARSE_ERROR);
	}
	arg[LINELEN] = '\0';

	if (strcmp(arg, "help") == 0) {
		shLog("OK dec atlimit, tolimit, raw NN, rel NN, abs NN");
	}
	else if (strcmp(arg, "atlimit") == 0) {
		/* 
		 * check the Dec limit switch
		 */
		retval = dec_is_at_limit();
		if (retval == 1) {
			shLog("OK dec atlimit yes");
		}
		else {
			shLog("OK dec atlimit no");
		}
	}
	else if (strcmp(arg, "raw") == 0) {
		/* 
		 * call the dec_move_raw routine
		 *   (only for debugging, not for user)
		 */
		if (narg != 3) {
			shError("dec raw: must supply number of steps");
			return(PARSE_ERROR);
		}
		retval = dec_move_raw(steps);
		if (retval == 0) {
			shLog("OK dec raw");
		}
		else {
			shError("dec raw");
		}
	}
	else if ((strcmp(arg, "tolimit") == 0) ||
	         (strcmp(arg, "home") == 0)) {
		/* 
		 * move the Dec motor so the telescope moves to
		 *   the Dec home position.
		 */
		retval = dec_move_to_limit();
		if (retval == 0) {
			shLog("OK dec tolimit");
		}
		else {
			shError("dec tolimit");
		}
	}
	else if (strcmp(arg, "rel") == 0) {
		if (narg != 3) {
			shError("dec rel: must supply number of steps");
			return(PARSE_ERROR);
		}
		retval = dec_move_rel(steps);
		if (retval == 0) {
			shLog("OK dec rel");
		}
		else {
			shError("dec rel");
		}
	}
	else if (strcmp(arg, "abs") == 0) {
		if (narg != 3) {
			shError("dec abs: must supply number of steps");
			return(PARSE_ERROR);
		}
		retval = dec_move_abs(steps);
		if (retval == 0) {
			shLog("OK dec abs");
		}
		else {
			shError("dec abs");
		}
	}
	else {
		shError("dec: unrecognized arg ..%s..", arg);
		return(PARSE_ERROR);
	}

	return(PARSE_OK);
}

	/*********************************************************************
	 * PROCEDURE: cmd_shutter
	 *
	 * DESCRIPTION: Handle the "shutter" command.
	 *
	 *              Usage:   shutter  [v|i]  [open|close]
	 *
	 *              Open or close one of the camera shutters.
	 *
	 *
	 * RETURNS:
	 *     PARSE_OK       if proper arguments given
	 *     PARSE_ERROR    if improper arguments
	 */

static int
cmd_shutter
	(
	char *cmdline              /* I: the entire command line, */
	                           /*    including the command word */
	)
{
	char cmd[LINELEN+1], camera[LINELEN+1], state[LINELEN+1];
	int narg, retval;

	narg = sscanf(cmdline, "%s %s %s", cmd, camera, state);
	if ((narg < 2) || (narg > 3)) {
		shError("cmd_shutter: wrong number of args");
		return(PARSE_ERROR);
	}
	camera[LINELEN] = '\0';
	state[LINELEN] = '\0';

	if (strcmp(camera, "help") == 0) {
		shLog("OK shutter [v|i] [open|close]");
	}
	else if (strcmp(camera, "v") == 0) {
		/* 
		 * open or close the V-band shutter
		 */
		if (strcmp(state, "open") == 0) {
			retval = servo(SERVO_VSHUTTER, SERVO_OPENSHUTTER);
			if (retval != 0) {
				shError("shutter: servo returns with error");
			}
			shLog("OK shutter v open");
		}
		else if (strcmp(state, "close") == 0) {
			retval = servo(SERVO_VSHUTTER, SERVO_CLOSESHUTTER);
			if (retval != 0) {
				shError("shutter: servo returns with error");
			}
			shLog("OK shutter v close");
		} 
		else {
			shError("shutter: invalid state %s", state);
			return(PARSE_ERROR);
		}
	}
	else if (strcmp(camera, "i") == 0) {
		/* 
		 * Open or close the I-band camera shutter
		 */
		if (strcmp(state, "open") == 0) {
			retval = servo(SERVO_ISHUTTER, SERVO_OPENSHUTTER);
			if (retval != 0) {
				shError("shutter: servo returns with error");
			}
			shLog("OK shutter i open");
		}
		else if (strcmp(state, "close") == 0) {
			retval = servo(SERVO_ISHUTTER, SERVO_CLOSESHUTTER);
			if (retval != 0) {
				shError("shutter: servo returns with error");
			}
			shLog("OK shutter i close");
		} 
		else {
			shError("shutter: invalid state %s", state);
			return(PARSE_ERROR);
		}
	}
	else {
		shError("shutter: unrecognized camera ..%s..", camera);
		return(PARSE_ERROR);
	}

	return(PARSE_OK);
}


	/*********************************************************************
	 * PROCEDURE: cmd_focus
	 *
	 * DESCRIPTION: Handle the "focus" command.
	 *
	 *              Usage:   focus    [v|i]  atlimit
	 *                                       tolimit  (or home)
	 *                                       abs NN
	 *                                       rel NN
	 *
	 *              Do something with one of the focus motors;
	 *              one must specify which one, the V-band
	 *              or I-band.  The options are:
	 *
	 *                   atlimit         tell if the given camera is
	 *                                      at its limit position
	 *
	 *                   tolimit         send the given camera to its 
	 *                   (or home)          limit position
	 *
	 *                   abs NN          move to limit, then NN steps
	 *                                      away from limit
	 *
	 *                   rel NN          move NN steps away from the
	 *                                      current position
	 *
	 *
	 * RETURNS:
	 *     PARSE_OK       if proper arguments given
	 *     PARSE_ERROR    if improper arguments
	 */

static int
cmd_focus
	(
	char *cmdline              /* I: the entire command line, */
	                           /*    including the command word */
	)
{
	char cmd[LINELEN+1], camera[LINELEN+1], arg[LINELEN+1];
	int narg = -1;
	int steps = -1;
	int retval;
	MOTOR_ID id;

	narg = sscanf(cmdline, "%s %s %s %d", cmd, camera, arg, &steps);
	if ((narg < 2) || (narg > 4)) {
		shError("cmd_focus: wrong number of args");
		return(PARSE_ERROR);
	}
	camera[LINELEN] = '\0';
	arg[LINELEN] = '\0';

	if (strcmp(camera, "help") == 0) {
		shLog("OK focus atlimit, tolimit, rel NN, abs NN");
		return(PARSE_OK);
	}

	/* 
	 * if we get here, the user must have supplied at least
	 *   2 args in addition to the "focus" command itself.
	 */
	if (narg < 3) {
		shError("cmd_focus: wrong number of args");
		return(PARSE_ERROR);
	}

	/* check to make sure that the first arg is a valid camera name */
	if (strcmp(camera, "v") == 0) {
		id = MOTOR_FOCUSV;
	}
	else if (strcmp(camera, "i") == 0) {
		id = MOTOR_FOCUSI;
	} 
	else {
		shError("cmd_focus: camera must be either v or i");
		return(PARSE_ERROR);
	}

	if (strcmp(arg, "atlimit") == 0) {
		/* 
		 * check the camera's limit switch
		 */
		retval = foc_is_at_limit(id);
		if (retval == 1) {
			shLog("OK focus %s atlimit yes ", camera);
		}
		else {
			shLog("OK focus %s atlimit no ", camera);
		}
	}
	else if ((strcmp(arg, "tolimit") == 0) ||
	         (strcmp(arg, "home") == 0)) {
		retval = foc_to_limit(id);
		if (retval != 0) {
			shError("focus %s tolimit", camera);
		}
		else {
			shLog("OK focus");
		}
	}
	else if (strcmp(arg, "rel") == 0) {
		if (narg != 4) {
			shError("focus rel: must supply number of steps");
			return(PARSE_ERROR);
		}
		retval = foc_move_rel(id, steps);
		if (retval != 0) {
			shError("focus %s rel", camera);
		}
		else {
			shLog("OK focus");
		}
	}
	else if (strcmp(arg, "abs") == 0) {
		if (narg != 4) {
			shError("focus abs: must supply number of steps");
			return(PARSE_ERROR);
		}
		retval = foc_move_abs(id, steps);
		if (retval != 0) {
			shError("focus %s abs", camera);
		}
		else {
			shLog("OK focus");
		}
	}
	else {
		shError("focus: unrecognized arg ..%s..", arg);
		return(PARSE_ERROR);
	}

	return(PARSE_OK);
}



	/*********************************************************************
	 * PROCEDURE: cmd_exp
	 *
	 * DESCRIPTION: Handle the "exp" command (which means "expose").
	 *
	 *              Usage:   exp  [dark|flat|object]  NN
	 *
	 *              Where
	 *                      dark         means take a dark image
	 *                                       (don't open shutters)
	 *
	 *                      flat         means take a flatfield image
	 *                                       (do open shutters)
	 *
	 *                      object       means take an image of an object
	 *                                       (do open shutters)
	 *
	 *                      NN           is the duration of the exposure
	 *                                       in seconds
	 *
	 *              Note that this command WILL save the the 
	 *              images to disk.
	 *
	 * RETURNS:
	 *     PARSE_OK       if proper arguments given
	 *     PARSE_ERROR    if improper arguments
	 */

static int
cmd_exp
	(
	char *cmdline              /* I: the entire command line, */
	                           /*    including the command word */
	)
{
	char cmd[LINELEN+1], arg[LINELEN+1];
	int narg = -1;
	int exptime = -1;
	double sleep_time = -1.0;

	narg = sscanf(cmdline, "%s %s %d", cmd, arg, &exptime);
	if ((narg < 2) || (narg > 3)) {
		shError("cmd_exp: wrong number of args");
		return(PARSE_ERROR);
	}
	arg[LINELEN] = '\0';

	if (strcmp(arg, "help") == 0) {
		shLog("OK exp [dark|flat|object] exptime ");
		return(PARSE_OK);
	}

	/* 
	 * if we get here, the user must have supplied at least
	 *   2 args in addition to the "exp" command itself.
	 */
	if (narg < 3) {
		shError("cmd_exp: wrong number of args");
		return(PARSE_ERROR);
	}
	if (exptime < 0) {
		shError("cmd_exp: exposure time must be >= 0");
		return(PARSE_ERROR);
	}

	/* check to make sure that the first arg is valid choice */
	if ((strcmp(arg, "dark") != 0) && 
	    (strcmp(arg, "flat") != 0) && 
	    (strcmp(arg, "object") != 0)) {
		shError("cmd_exp: arg must be 'dark', 'flat', or 'object'");
		return(PARSE_ERROR);
	}
	set_exp_type(arg);

	/* 
	 * check to make sure that the exposure time is a valid choice:
	 *   it must be >= 0 seconds.  For non-dark exposures, the
	 *   time for shutters to open and close yields a minimum
	 *   possible exposure time of EXPOSURE_BUILTIN seconds.
	 *   So, for non-dark exposures, we calculate the difference
	 *        
	 *      sleep_time = requested exposure time - EXPOSURE_BUILTIN
	 *
	 *   and we'll sleep for that amount of time.
	 *
	 *   If requested non-dark exposure time is less than 
	 *   EXPOSURE_BUILTIN, then print an error message and
	 *   return with error.
	 */
	if (exptime < 0) {
		shError("cmd_exp: exposure time must be >= 0 seconds");
		return(PARSE_ERROR);
	}
	/* note that sleep_time is a double, not an int .... */
	sleep_time = (double) exptime;
	if (strcmp(arg, "dark") != 0) {
		sleep_time -= EXPOSURE_BUILTIN;
		if (sleep_time < 0.0) {
			shError("cmd_exp: min non-dark exptime is %.1f",
					EXPOSURE_BUILTIN);
			return(PARSE_ERROR);
		}
	}


	/* 
	 * the first step is always to close the shutters
	 *   (if they are already closed, this does nothing 
	 */
#ifdef DEBUG
	shLog("    cmd_exp: about to close both shutters");
#endif
	if (servo(SERVO_ISHUTTER, SERVO_CLOSESHUTTER) != 0) {
		shError("cmd_exp: servo fails to close I shutter");
	}
	if (servo(SERVO_VSHUTTER, SERVO_CLOSESHUTTER) != 0) {
		shError("cmd_exp: servo fails to close V shutter");
	}

	/* 
	 * the second step is always to make a clearing scan
	 */
#ifdef DEBUG
	shLog("    cmd_exp: about to start clearing scan");
#endif
	if (scan_both() != 0) {
		shError("cmd_exp: scan_both fails on clearing scan");
	}
	
	/* 
	 * save the current time for future reference (we need it
	 *   to construct FITS header).  Ditto the exposure time.
	 */
	set_exp_length(exptime);
	set_exp_clock();

	/*
	 * now, if this is not a "dark" exposure, we open shutters
	 */
	if (strcmp(arg, "dark") != 0) {
#ifdef DEBUG
		shLog("    cmd_exp: about to open both shutters");
#endif
		if (servo(SERVO_ISHUTTER, SERVO_OPENSHUTTER) != 0) {
			shError("cmd_exp: servo fails to open I shutter");
		}
		if (servo(SERVO_VSHUTTER, SERVO_OPENSHUTTER) != 0) {
			shError("cmd_exp: servo fails to open V shutter");
		}
	}

	/* 
	 * wait for the given exposure time 
	 */
#ifdef DEBUG
	shLog("    cmd_exp: about to sleep for %.1f seconds", sleep_time);
#endif
	usleep((int) (sleep_time*1000000));

	/* 
	 * now, if this was not a "dark" exposure, close shutters 
	 */
	if (strcmp(arg, "dark") != 0) {
#ifdef DEBUG
		shLog("    cmd_exp: about to close both shutters");
#endif
		if (servo(SERVO_ISHUTTER, SERVO_CLOSESHUTTER) != 0) {
			shError("cmd_exp: servo fails to close I shutter");
		}
		if (servo(SERVO_VSHUTTER, SERVO_CLOSESHUTTER) != 0) {
			shError("cmd_exp: servo fails to close V shutter");
		}
	}

	/* 
	 * the final step is always to make a readout scan
	 */
#ifdef DEBUG
	shLog("    cmd_exp: about to start readout scan");
#endif
	if (scan_both() != 0) {
		shError("cmd_exp: scan_both fails on readout scan");
	}

	/* 
	 * now, save the images to disk
	 */
	if (save_images() != 0) {
		shError("cmd_exp: save_images fails");
	}

	shLog("OK exp");

	return(PARSE_OK);
}

	/*********************************************************************
	 * PROCEDURE: cmd_show
	 *
	 * DESCRIPTION: Handle the "show" command 
	 *
	 *              Usage:  show  adc
	 *                            dac 
	 *                            seq
	 *
	 *              Where
	 *                      adc          means to print out all the
	 *                                      ADC values
	 *
	 *                      dac          means to print out all the
	 *                                      DAC values
	 *
	 *                      seq          means to print out the current
	 *                                      image sequence number
	 *
	 * RETURNS:
	 *     PARSE_OK       if proper arguments given
	 *     PARSE_ERROR    if improper arguments
	 */

static int
cmd_show
	(
	char *cmdline              /* I: the entire command line, */
	                           /*    including the command word */
	)
{
	char cmd[LINELEN+1], arg[LINELEN+1];
	char line[LINELEN+1];
	int narg = -1;
	int seq = -1;

	narg = sscanf(cmdline, "%s %s", cmd, arg);
	if ((narg < 2) || (narg > 2)) {
		shError("cmd_show: wrong number of args");
		return(PARSE_ERROR);
	}
	arg[LINELEN] = '\0';

	if (strcmp(arg, "help") == 0) {
		shLog("OK show adc, dac, seq");
		return(PARSE_OK);
	}

	if (strcmp(arg, "adc") == 0) {
		show_adcs();
	}
	else if (strcmp(arg, "dac") == 0) {
		show_dacs();
	}
	else if (strcmp(arg, "seq") == 0) {
		seq = get_sequence();
		sprintf(line, "seq %03d ", seq);
		shLog(line);
	}
	else {
		shError("show: unrecognized arg ..%s..", arg);
		return(PARSE_ERROR);
	}

	shLog("OK show");
	return(PARSE_OK);
}

	/*********************************************************************
	 * PROCEDURE: cmd_set
	 *
	 * DESCRIPTION: Handle the "set" command 
	 *
	 *              Usage:  set  dac NN xxxx
	 *                           seq NN
	 *
	 *
	 *              Where
	 *                      dac         indicates that we are to set a 
	 *                                      single DAC value
	 *                      NN          is the index of the DAC to set
	 *                      xxxx        is the voltage to set 
	 *                                      (floating-point value)
	 *
	 *                      seq         indicates that we are to set 
	 *                                      the image sequence value
	 *                      NN          is the desired sequence value
	 *
	 * RETURNS:
	 *     PARSE_OK       if proper arguments given
	 *     PARSE_ERROR    if improper arguments
	 */

static int
cmd_set
	(
	char *cmdline              /* I: the entire command line, */
	                           /*    including the command word */
	)
{
	char cmd[LINELEN+1], arg[LINELEN+1];
	int narg = -1;
	int index;
	double value;

	narg = sscanf(cmdline, "%s %s %d %lf", cmd, arg, &index, &value);
	if ((narg < 2) || (narg > 4)) {
		shError("cmd_set: wrong number of args");
		return(PARSE_ERROR);
	}
	arg[LINELEN] = '\0';

	if (strcmp(arg, "help") == 0) {
		shLog("OK set dac NN xxxx, seq NN");
		return(PARSE_OK);
	}

	/*
	 * handle the case of "set dac NN xxxx" 
	 *   which means "set the DAC with index NN to the given value xxxx" 
	 */
	if ((narg == 4) && (strcmp(arg, "dac") == 0)) {
		if (load_one_dac(index, value) != 0) {
			shError("cmd_set: set_one_dac returns with error");
		}
	}
	/* 
	 * handle the case of "set seq NN", which sets the image 
	 *   sequence number for the current run to the third argument
	 *   (stored in "index")
	 */
	else if ((narg == 3) && (strcmp(arg, "seq") == 0)) {
		if (set_sequence(index) != 0) {
			shError("cmd_set: set_sequence returns with error");
		}
	}
	else {
		shError("set: unrecognized command string");
		return(PARSE_ERROR);
	}

	shLog("OK set");
	return(PARSE_OK);
}



	/*********************************************************************
	 * PROCEDURE: cmd_saveimg
	 *
	 * DESCRIPTION: Handle the "saveimg" command 
	 *
	 *              Usage:  saveimg
	 *
	 *              Currently, this calls an external program
	 *              to read out the images from the memory card
	 *              and save them in FITS files on disk.
	 *
	 * RETURNS:
	 *     PARSE_OK       if proper arguments given
	 *     PARSE_ERROR    if improper arguments
	 */

static int
cmd_saveimg
	(
	char *cmdline              /* I: the entire command line, */
	                           /*    including the command word */
	)
{
	char cmd[LINELEN+1], arg[LINELEN+1];
	int narg = -1;

	narg = sscanf(cmdline, "%s %s", cmd, arg);
	if (narg > 2) {
		shError("cmd_saveimg: wrong number of args");
		return(PARSE_ERROR);
	}
	arg[LINELEN] = '\0';

	if (strcmp(arg, "help") == 0) {
		shLog("OK saveimg");
		return(PARSE_OK);
	}

	if (save_images() != 0) {
		shError("cmd_saveimg: save_images returns with error");
	}

	return(PARSE_OK);
}



	/*********************************************************************
	 * PROCEDURE: cmd_saveimg
	 *
	 * DESCRIPTION: Handle the "radec" command 
	 *
	 *              Usage:  radec ra dec
	 *
	 *              Move the telescope to the given (RA, Dec) position.
	 *              Both "ra" and "dec" arguments must be in decimal degrees.
	 *
	 * RETURNS:
	 *     PARSE_OK       if proper arguments given
	 *     PARSE_ERROR    if improper arguments
	 */

static int
cmd_radec
	(
	char *cmdline              /* I: the entire command line, */
	                           /*    including the command word */
	)
{
	char cmd[LINELEN+1], arg1[LINELEN+1], arg2[LINELEN+1];
	int narg = -1;
	int retval;
	double ra, dec;

	narg = sscanf(cmdline, "%s %s %s", cmd, arg1, arg2);
	if (narg != 3) {
		shError("cmd_saveimg: wrong number of args");
		return(PARSE_ERROR);
	}
	arg1[LINELEN] = '\0';

	if (strcmp(arg1, "help") == 0) {
		shLog("OK radec ra dec");
		return(PARSE_OK);
	}

	/* get the desired RA and Dec */
	if (sscanf(arg1, "%lf", &ra) != 1) {
		shError("cmd_radec: cannot read RA in decimal degrees from arg %s", 
						arg1);
		return(PARSE_ERROR);
	}
	if (sscanf(arg2, "%lf", &dec) != 1) {
		shError("cmd_radec: cannot read Dec in decimal degrees from arg %s", 
						arg2);
		return(PARSE_ERROR);
	}

	/* call the "goto_radec" function and return the appropriate code */
	retval = goto_radec(ra, dec);
	if (retval == RADEC_INVALID) {
		shError("cmd_radec: given RA and/or Dec was out of range");
		return(PARSE_OK);
	}
	if (retval == RADEC_ERROR) {
		shError("cmd_radec: an error occurred while trying to move telescope");
		return(PARSE_OK);
	}

	return(PARSE_OK);
}






/* 
 * low-level Dec motor commands 
 */


	/**********************************************************
	 * PROCEDURE: dec_move_raw
	 *
	 * DESCRIPTION: Move the Dec motor by the given number
	 *              of steps, relative to its current 
	 *              position.  
	 *
	 *              The argument 'steps' tells us how many steps to move.
	 *                 positive    ->  move North
	 *                 negative    ->  move South
	 *
	 *              Do NOT unclamp/clamp the Dec brake!
	 *
	 *              This routine should never be called by
	 *              the user, only by other routines
	 *              (or for debugging).
	 * 
	 * RETURNS:
	 *       0            if all goes well
	 *       1            if there's an error
	 */

static int 
dec_move_raw
	(
	int steps             /* I: number of steps to move */
	)
{
	int retval;
	MOTOR_DIRECTION dir;

	/* if asked to move zero steps, just return immediately */
	if (steps == 0) {
		return(0);
	}

	if (steps > 0) {
		/* away from the limit = North */
		dir = MOTOR_FROM_LIMIT;
	}
	else {
		/* in towards the limit = South */
		dir = MOTOR_TOWARD_LIMIT;
		steps = 0 - steps;
	}

	retval = motor(MOTOR_IN_DIRECTION, MOTOR_DECDRIVE, dir, steps);
	if (retval != 0) {
		shError("dec_move_raw: motor returns with error");
	}

	return(retval);
}


	/**********************************************************
	 * PROCEDURE: dec_move_to_limit
	 *
	 * DESCRIPTION: Move the Dec motor so that it reaches
	 *              its limit position on the Northern side.
	 *
	 * Move the telescope so that it reaches the Dec limit 
	 *   (on Northern side).
	 *   We should be able to use the "Action=0" option of 
	 *   the Motor subroutine,
	 *   but it is not working.  So, we have to make series of moves
	 *   and check the status of the Dec limit switch repeatedly.
	 *
	 * This routine will always cause Dec motion, though we try to make
	 *   very small motions at first to see if we're already very close
	 *   to the home position.
	 *
	 * This routine does unclamp/clamp the Dec motor.
	 * 
	 * RETURNS:
	 *       0            if all goes well
	 *       1            if there's an error
	 */

static int 
dec_move_to_limit
	(
	void
	)
{
	int limit_pos;
	/* 
	 * these are the amounts by which we command Dec motor to move
	 *   The units are "stepper motor steps"; these values work
	 *   well for me.  MWR 8/13/2001.
	 */
	int small_step = 10;
	int big_step = 100;
	
	/* first, we loosen the Dec clamp */
	if (cmd_clamp("clamp off") != 0) {
		shError("dec_move_to_limit: cmd_clamp off returns with error");
		return(1);
	}

	/* 
	 * check to see if the Dec limit switch is lit; if so, then
	 *   we follow one set of rules.  If not, we follow a different
	 *   set....
	 */
	if (dec_is_at_limit() == 1) {
		shLog("   dec_move_to_limit: limit on");
		
		/* 
		 * maybe we are very close to (or at) the home 
		 *   position right now.  If so, then moving a tiny
		 *   bit South will cause the limit switch to
		 *   go dark.  Check for this ....
		 */
		dec_move_raw(-small_step);
		if (dec_is_at_limit() == 0) {
			shLog("    moving teeny bit South passed limit");
			
			/* 
			 * so, just move back a bit to the North, and 
			 *   we'll be at limit again.
			 */
			dec_move_raw(small_step);
		}
		else {

			/* 
		 	 * if we get here, we must be far North of the limit
		 	 *   switch.  So, let's move South by big steps until
		 	 *   we trigger it.  Then, we'll move back North by
		 	 *   small steps until we pass it again going 
			 *   Northwards.
		 	 */
			shLog("    moving big steps South to pass limit");

			while (dec_is_at_limit() == 1) {
				dec_move_raw(-big_step);
			}
			shLog("    back up by tiny steps North to pass limit");
			while (dec_is_at_limit() == 0) {
				dec_move_raw(small_step);
			}

		}

	}
	else {
		/* 
		 * in this case, we started with the Dec limit switch
		 *   dark, which means that we need to move North to 
		 *   reach the home position.
		 */
		shLog("    dec_move_to_limit: limit off");


		/* 
		 * Maybe we are very close to the home position.
		 *   If so, then moving just a small amount North with
		 *   cause the limit switch to trigger.
		 */
		dec_move_raw(small_step);
		if (dec_is_at_limit() == 1) {
			shLog("    moving small step North reached limit");
			/* and we are done */
		}
		else {
			/*
			 * nope, we must be pretty far South of the limit.
			 *   So, we move North by big steps until we pass
			 *   the limit, then back up with small steps 
			 *   until we pass it again.
			 */
			shLog("    small amount North failed to reach limit");

			while (dec_is_at_limit() == 0) {
				dec_move_raw(big_step);
			}
			shLog("    back up small steps to pass limit again");
			while (dec_is_at_limit() == 1) {
				dec_move_raw(-small_step);
			}
			dec_move_raw(small_step);

		}
	}

	limit_pos = dec_is_at_limit();
	shLog("    at end of dec_move_to_limit, switch is %d", limit_pos);

	/* finally, we tighten the Dec clamp */
	if (cmd_clamp("clamp on") != 0) {
		shError("dec_move_to_limit: cmd_clamp on returns with error");
		return(1);
	}

  	return(0);
}


	/**********************************************************
	 * PROCEDURE: dec_move_rel
	 *
	 * DESCRIPTION: Unclamp the Dec motor,
	 *              move the Dec axis by the given number
	 *              of steps, relative to its current position,
	 *              re-clamp the Dec motor.
	 *
	 *              Ordinary users should call this routine,
	 *              not "dec_move_raw".
	 *
	 *              The argument 'steps' tells us how many steps to move.
	 *                 positive    ->  move North
	 *                 negative    ->  move South
	 *
	 * RETURNS:
	 *       0            if all goes well
	 *       1            if there's an error
	 */

static int 
dec_move_rel
	(
	int steps             /* I: number of steps to move */
	)
{
	int retval;

	/* start by loosening the Dec clamp */
	if (cmd_clamp("clamp off") != 0) {
		shError("dec_move_rel: cmd_clamp off returns with error");
		return(1);
	}

	retval = dec_move_raw(steps);
	if (retval != 0) {
		shError("dec_move_rel: dec_move_raw returns with error");
	}

	/* we tighten the Dec clamp */
	if (cmd_clamp("clamp on") != 0) {
		shError("dec_move_rel: cmd_clamp on returns with error");
		return(1);
	}

	return(retval);
}


	/**********************************************************
	 * PROCEDURE: dec_move_abs
	 *
	 * DESCRIPTION: Unclamp the Dec motor,
	 *              go to the Dec home position,
	 *              move the given number of steps away from home,
	 *              and finally re-clamp the Dec motor.
	 *
	 *              The argument 'steps' tells us how many steps to move.
	 *                 positive    ->  move North
	 *                 negative    ->  move South
	 *
	 *              Note that we first call "dec_move_to_limit"
	 *              (which unclamps and re-clamps),
	 *              then move to the desired position --
	 *              which involves a second round of unclamp
	 *              and re-clamp.  This is somewhat inefficient,
	 *              but the code is tighter.  Will probably
	 *              pay off in lower maintenance later.
	 *
	 * RETURNS:
	 *       0            if all goes well
	 *       1            if there's an error
	 */

static int 
dec_move_abs
	(
	int steps             /* I: number of steps to move away from home */
	)
{
	int retval;

	/* start by going to the Dec home position. */
	if (dec_move_to_limit() != 0) {
		shError("dec_move_abs: dec_move_to_limit fails");
		return(1);
	}

	/*
	 * now, loosen the Dec clamp (which was just loosened and
	 *   re-tightened by the dec_move_to_limit routine above) 
	 */
	if (cmd_clamp("clamp off") != 0) {
		shError("dec_move_abs: cmd_clamp off returns with error");
		return(1);
	}

	retval = dec_move_raw(steps);
	if (retval != 0) {
		shError("dec_move_abs: dec_move_raw returns with error");
	}

	/* finish: tighten the Dec clamp */
	if (cmd_clamp("clamp on") != 0) {
		shError("dec_move_abs: cmd_clamp on returns with error");
		return(1);
	}

	return(retval);
}


/*
 * low-level camera focus commands
 */

	/**********************************************************
	 * PROCEDURE: foc_to_limit
	 *
	 * DESCRIPTION: Move the given camera to the focus limit
	 *              position, as close to the lens as possible.
	 *
	 * RETURNS:
	 *       0            if all goes well
	 *       1            if there's an error
	 */

static int 
foc_to_limit
	(
	MOTOR_ID id           /* I: which camera's motor? */
	)
{
	int retval;
	int steps;
	MOTOR_ACTION action;
	MOTOR_DIRECTION direction;

	/* if we're already at limit, return immediately */
	if (foc_is_at_limit(id) == 1) {
#ifdef DEBUG
		shLog("    foc_to_limit: already at limit");
#endif
		return(0);
	}

	/* tell the 'motor' function to move the appropriate motor to limit */
	action = MOTOR_LIMIT_OUT;
	direction = MOTOR_TOWARD_LIMIT;
	steps = 0;
	retval = motor(action, id, direction, steps);
	if (retval != 0) {
		shError("foc_to_limit: motor returns with error");
		return(1);
	}

	return(retval);
}


	/**********************************************************
	 * PROCEDURE: foc_move_rel
	 *
	 * DESCRIPTION: Move the given camera relative to 
	 *              its current position by the given number
	 *              of steps.
	 *
	 *              steps > 0       move away from limit switch
	 *                              push camera away from sky
	 *
	 *              steps < 0       move towards limit switch
	 *                              push camera towards the sky
	 *
	 * RETURNS:
	 *       0            if all goes well
	 *       1            if there's an error
	 */

static int 
foc_move_rel
	(
	MOTOR_ID id,          /* I: move this camera's motor */
	int steps             /* I: move this many steps */
	)
{
	int retval;
	MOTOR_ACTION action;
	MOTOR_DIRECTION direction;

	/* move the given number of steps away from limit */
	action = MOTOR_IN_DIRECTION;
	if (steps > 0) {
		direction = MOTOR_FROM_LIMIT;
	} else {
		direction = MOTOR_TOWARD_LIMIT;
		steps = 0 - steps;
	}
	retval = motor(action, id, direction, steps);
	if (retval != 0) {
		shError("foc_move_rel: motor returns with error");
		return(1);
	}

	return(retval);
}


	/**********************************************************
	 * PROCEDURE: foc_move_abs
	 *
	 * DESCRIPTION: Move the given camera to the focus limit
	 *              position, then back out by the given
	 *              number of steps.
	 *
	 *              The given number of steps must be positive,
	 *              or else we return with an error code.
	 *
	 * RETURNS:
	 *       0            if all goes well
	 *       1            if there's an error
	 */

static int 
foc_move_abs
	(
	MOTOR_ID id,          /* I: move this camera's motor */
	int steps             /* I: move this many steps away from limit */
	)
{

	/* make sure that 'steps' is positive */
	if (steps < 0) {
		shError("foc_move_abs: steps must be positive");
		return(1);
	}

	/* first, we move to the limit switch */
	if (foc_to_limit(id) != 0) {
		shError("foc_move_abs: foc_to_limit fails");
		return(1);
	}

	/* now, move the given number of steps away from limit */
	if (foc_move_rel(id, steps) != 0) {
		shError("foc_move_abs: foc_move_rel fails");
	 	return(1);
	}

	return(0);
}


/*
 * mid-level telescope motion commands
 */


	/***********************************************************************
	 * PROCEDURE: goto_radec
	 *
	 * DESCRIPTION: The user provides an (RA, Dec), in decimal degrees.
	 *              We do the following:
	 *
	 *                 1. Check to make sure that it lies within the 
	 *                    acceptable range of motion.  If not, 
	 *                    we return with code 1.
	 *
	 *                 2. Go to Dec home, then the given Dec
	 *
	 *                 3. Go to RA home, then the given RA
	 *
	 * RETURNS:
	 *    RADEC_OK       if all goes well
	 *    RADEC_INVALID  if the given (RA, Dec) position in not currently valid
	 *    RADEC_ERROR    if some error occurred
	 */

static int 
goto_radec
	(
	double ra,              /* I: desired RA, in decimal degrees */
	double dec              /* I: desired Dec, in decimal degrees */
	)
{
	int steps;
	int seconds;
	double ha;

	/* 
	 * step 1: calculate the Hour Angle for the given RA --
	 */
	get_ha(ra, &ha);

	/* step 2: check that the given (RA, Dec) in within valid ranges */
	if ((ha < HA_LIMIT_MIN) || (ha > HA_LIMIT_MAX)) {
		shError("ha %9.4f is outside permitted range", ha);
		return(RADEC_INVALID);
	}
	if ((dec < DEC_LIMIT_MIN) || (dec > DEC_LIMIT_MAX)) {
		shError("dec %9.4f is outside permitted range", dec);
		return(RADEC_INVALID);
	}


	/* 
	 * step 3: move to the desired Dec.
	 */
	steps = floor((dec - DEC_OF_HOME)/DEC_STEP_SIZE);
#ifdef DEBUG
	shLog("goto_radec: Dec %7.2f corresponds to %6d steps", dec, steps);
#endif
	if (dec_move_abs(steps) != 0) {
		return(RADEC_ERROR);
	}

	/*
	 * step 4: move to the desired RA (really, the desired Hour Angle)
	 */
	seconds = floor((ha - HA_OF_HOME)/(RA_SLEW_RATE));
#ifdef DEBUG
	shLog("goto_radec: RA %7.2f = HA %7.2f means %6d seconds", ra, ha, seconds);
#endif
	if (cmd_ra("ra tolimit") != PARSE_OK) {
		shError("goto_radec: cmd_ra fails on 'ra tolimit'");
		return(RADEC_ERROR);
	}
	if (cmd_ra("ra rew") != PARSE_OK) {
		shError("goto_radec: cmd_ra fails on 'ra rew'");
		return(RADEC_ERROR);
	}
	/* now wait for the proper number of seconds while RA drive moves */
	sleep(seconds);
	/* stop the RA drive */
	if (cmd_ra("ra off") != PARSE_OK) {
		shError("goto_radec: cmd_ra fails on 'ra off'");
		return(RADEC_ERROR);
	}


	return(RADEC_OK);
}

