
	/*
	 * Code to read information from a file, one line at a time,
	 *   in such a way that the file can be modified in between
	 *   readings.
	 *
	 * MWR 8/27/2001
	 */

#include <stdio.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <string.h>
#include <unistd.h>
#include <errno.h>
#include "tait.h"

#define TEMPNAME "./flock.tmp"


static int lock(char *filename);
static int unlock(char *filename);

#ifdef TESTMAIN
main(int argc, char *argv[])
{
	char filename[LINELEN+1];
	char first_line[LINELEN+1];

	strcpy(filename, "./foo");

	while (get_next_line(filename, first_line) == 0) {
		if (first_line[0] == '\0') {
			break;
		}
		printf("next line is ..%s..\n", first_line);
		printf("  hit return to continue: ");
		fgets(first_line, LINELEN, stdin);
	}


	exit(0);
}
#endif


	/*****************************************************************
	 * PROCEDURE: get_next_line
	 *
	 * DESCRIPTION: Gets the next line from the given file, in the
	 *              following manner:
	 *
	 *                 1. lock file so no one can write to it
	 *                 2. read the first line from file
	 *                        (if file is empty, place '\0' in first char
	 *                         of the 'first_line' argument)
	 *                        save the first line for later use
	 *                 3. write all remaining lines into a temp file
	 *                 4. close input and temp file
	 *                 5. unlock file
	 *                 6. rename temp file to input file
	 *
	 *              The first line we read is placed into the
	 *              "first_line" argument passed to the function.
	 *              That argument must point to a string of at
	 *              least LINELEN characters.
	 *
	 *              The goal is to set things up so that we can read
	 *              the file sequentially, but permit the possibility
	 *              that someone may modify the file between our reads.
	 *
	 * RETURNS:
	 *    0         if we were able to read a line of information
	 *    1         if some error occurs, or we reach EOF
	 */


int
get_next_line
	(
	char *filename,          /* I: name of file to read */
	char *first_line         /* I: place the first line into this arg */
	                         /*      or '\0' if file is empty */
	)
{
	char tempfile[LINELEN+1];
	char line[LINELEN+1];
	FILE *input_fp, *temp_fp;

	strcpy(tempfile, TEMPNAME);

	/* step 1: lock the input file so no one can write to it */
	lock(filename);

	/* step 1A: open the input file (reading) and temp file (writing) */
	if ((input_fp = fopen(filename, "r")) == NULL) {
		fprintf(stderr, "get_next_line: can't open %s for input", 
						filename);
		return(1);
	}
	if ((temp_fp = fopen(tempfile, "w")) == NULL) {
		fprintf(stderr, "get_next_line: can't open %s for output", 
						tempfile);
		return(1);
	}

	/* step 2: read the first line from input file */
	if (fgets(first_line, LINELEN, input_fp) == NULL) {
		/* input file is empty ... so, we're done */
		first_line[0] = '\0';
		fclose(input_fp);
		fclose(temp_fp);
		unlink(tempfile);
		unlock(filename);
		return(1);
	}

	/* step 3: copy all OTHER lines into the temp file */
	while (fgets(line, LINELEN, input_fp) != NULL) {
		fputs(line, temp_fp);
	}

	/* step 4: close input and temp files */
	fclose(input_fp);
	fclose(temp_fp);

	/* step 5: unlock file so we can rename it */
	unlock(filename);

	/* step 6: rename the temp file to the input file */
	if (rename(tempfile, filename) != 0) {
		perror(sys_errlist[errno]);
		fprintf(stderr, "rename fails for old %s new %s \n", tempfile, filename);
		exit(1);
	}

	/* step 7: unlock the file again (which sets permissions properly) */
	unlock(filename);

	return(0);
}





static int
lock(char *filename)
{
	int retval;

	retval = chmod(filename, S_IRUSR);
	if (retval != 0) {
		perror(sys_errlist[errno]);
		return(1);
	}

	return(0);
}


static int
unlock(char *filename)
{
	int retval;

	retval = chmod(filename, S_IRUSR | S_IWUSR | 
	                         S_IRGRP | S_IWGRP |
				 S_IROTH | S_IWOTH);
	if (retval != 0) {
		perror(sys_errlist[errno]);
		return(1);
	}

	return(0);
}

