
	/*
	 * This file contains routines which save an image to
	 * a disk FITS file.
	 */

#include <stdio.h>
#include <unistd.h>
#include <string.h>
#include <ctype.h>
#include <time.h>
#include <math.h>
#include "tait.h"

	
	/* keeps track of images taken during a run (night) */
static int sequence_number;
	/* the system time when we started an observation */
static time_t exp_clock;
	/* the length of the most recent observation (seconds) */
static int exp_length;
	/* the type of the most recent observation */
static char exp_type[LINELEN+1];

	/* local static functions */
static void incr_sequence(void);
static time_t get_exp_clock(void);
static int get_exp_length(void);
static char *get_exp_type(void);
static int set_hdr_keywords(char *filename, char *hdr_dateobs,
                            char *hdr_ut, char *hdr_exptime, 
			    char *hdr_imagetyp);



	/****************************************************************
	 * PROCEDURE: init_sequence
	 *
	 * DESCRIPTION: Initialize to zero the running sequence number
	 *              for images taken during a night.
	 *
	 *              Obviously, this should be called only once
	 *              per night.  Otherwise, new images will overwrite
	 *              old ones.
	 *
	 * RETURNS:
	 *        nothing
	 */

void
init_sequence
	(
	void
	)
{
	set_sequence(0);
}

	/****************************************************************
	 * PROCEDURE: set_sequence
	 *
	 * DESCRIPTION: Set the sequence number for images taken 
	 *              during a night to the given value.
	 *
	 * RETURNS:
	 *    0         if all goes well
	 *    1         if there's a problem
	 */

int
set_sequence
	(
	int sequence_value        /* I: set sequence number to this value */
	)
{
	if ((sequence_value < 0) || (sequence_value > 999)) {
		shError("   set_sequence: value %d is outside range 0 - 999",
				sequence_value);
		return(1);
	}

	sequence_number = sequence_value;
	return(0);
}

	/****************************************************************
	 * PROCEDURE: get_sequence
	 *
	 * DESCRIPTION: Return the current sequence number, used for
	 *              keeping track of images in a run.
	 *
	 * RETURNS:
	 *      the current sequence number
	 */

int
get_sequence
	(
	void
	)
{
	return(sequence_number);
}


	/****************************************************************
	 * PROCEDURE: incr_sequence
	 *
	 * DESCRIPTION: Increment the current sequence number, used for
	 *              keeping track of images in a run.
	 *
	 * RETURNS:
	 *      nothing
	 */

static void
incr_sequence
	(
	void
	)
{
	sequence_number++;
}


	

	/****************************************************************
	 * PROCEDURE: generate_name
	 *
	 * DESCRIPTION: This routine generates a file name template,
	 *              which will form the basis of the file names
	 *              into which the current images will be saved.
	 *              The template has the form
	 *
	 *                    jjjjxxxc.fit
	 *
	 *              where
	 *                     jjjj         is the integer portion of
	 *                                    Julian Date - 2,450,000
	 *
	 *                      xxx         is an image sequence number 
	 *                                    within the current run
	 *                                    (typically one run per night)
	 *
	 *                      c           is a code indicating the camera
	 *                                      'E' for the V-band camera
	 *                                      'F' for the I-band camera
	 *
	 *              Actually, this routine leaves the "c" character
	 *              set to 'x'.  It will be changed to the appropriate
	 *              value when an image is actually saved to disk.
	 *
	 *              This function has to query information about
	 *              the image (exposure start, exposure length,
	 *              exposure type) in order to modify the FITS 
	 *              header values in the "template" files.
	 *
	 *              This function modifies the FITS header
	 *              "template" files, CCD_PLATEI and CCD_PLATEV,
	 *              modifying several of the FITS keywords so that
	 *              they are appropriate for the latest observation.
	 *
	 *
	 * RETURNS:
	 *     0            if all goes well
	 *     1            if an error occurs
	 */

int
generate_name
	(
	char *name               /* O: a char array at least IMAGENAME */
	                         /*       characters long.  This routine */
	                         /*       will change the values of the */
				 /*       elements of the array */
	)
{
	char hdr_dateobs[LINELEN+1];
	char hdr_ut[LINELEN+1];
	char hdr_exptime[LINELEN+1];
	char hdr_imagetyp[LINELEN+1];
	char exptype[LINELEN+1];
	int year, month, day;
	int seq, jd_trunc;
	double ut_day, jd;
	double exptime;
	time_t start_time, this_time;
	struct tm *tms;

	shAssert(name != NULL);

	/* get the saved time at the start of the exposure */
	start_time = get_exp_clock();
	/* get the length (in seconds) of the exposure */
	exptime = (double) get_exp_length();
	/* get the type of the exposure */
	strncpy(exptype, get_exp_type(), LINELEN);

	/* 
	 * for both I and V exposures, we add the appropriate delays
	 * to the "start_time", then convert to UT.  
	 *
	 * Note that a "dark" exposure does not have any 
	 * shutter-induced delays; all other types are assumed to 
	 * wait for the shutters.
	 */

	/* first, I-band */
	if (strcmp(exptype, "dark") == 0) {
		this_time = floor((double) start_time 
			           + (exptime/2.0) + 0.5);
	} 
	else {
		this_time = floor((double) start_time + EXPOSURE_IDELAY 
			           + (exptime/2.0) + 0.5);
	}
	tms = gmtime(&this_time);
	year = tms->tm_year + 1900;
	month = tms->tm_mon + 1;
	day = tms->tm_mday;
	sprintf(hdr_dateobs, "DATE-OBS= '%04d-%02d-%02d'", year, month, day);
	sprintf(hdr_ut, "UT      = '%02d:%02d:%02d'", 
			tms->tm_hour, tms->tm_min, tms->tm_sec);
	sprintf(hdr_exptime, "EXPTIME = %.1f ", exptime);
	sprintf(hdr_imagetyp, "IMAGETYP= '%s' ", exptype);
	
	/*
	 * set FITS header values in the I-band template file 
	 */
	if (set_hdr_keywords(CCD_PLATEI, hdr_dateobs, 
				hdr_ut, hdr_exptime, hdr_imagetyp) != 0) {
		shError("generate_name: set_hdr_keywords fails for file %s",
				CCD_PLATEI);
		return(1);
	}


	/* now, calculate mid-exposure time in V-band */
	if (strcmp(exptype, "dark") == 0) {
		this_time = floor((double) start_time 
			           + (exptime/2.0) + 0.5);
	} 
	else {
		this_time = floor((double) start_time + EXPOSURE_VDELAY 
			           + (exptime/2.0) + 0.5);
	}
	tms = gmtime(&this_time);
	year = tms->tm_year + 1900;
	month = tms->tm_mon + 1;
	day = tms->tm_mday;
	sprintf(hdr_dateobs, "DATE-OBS= '%04d-%02d-%02d'", year, month, day);
	sprintf(hdr_ut, "UT      = '%02d:%02d:%02d'", 
			tms->tm_hour, tms->tm_min, tms->tm_sec);
	sprintf(hdr_exptime, "EXPTIME = %.1f ", exptime);
	sprintf(hdr_imagetyp, "IMAGETYP= '%s' ", exptype);
	
	/*
	 * set FITS header values in the V-band template file 
	 */
	if (set_hdr_keywords(CCD_PLATEV, hdr_dateobs, 
				hdr_ut, hdr_exptime, hdr_imagetyp) != 0) {
		shError("generate_name: set_hdr_keywords fails for file %s",
				CCD_PLATEV);
		return(1);
	}


	/* 
	 * now get the Julian Date for the date and time at mid-exposure 
	 *   (of the second image, i.e. V-band) 
	 */
	ut_day = tms->tm_hour + (tms->tm_min/60.0) + (tms->tm_sec/86400.0);
	get_jd(day, tms->tm_mon, year, ut_day, &jd);
	jd_trunc = floor(jd - 2450000.0);

	/* get the current sequence number */
	seq = get_sequence();

	/* build the template name */
	sprintf(name, "%04d%03dx.fit", jd_trunc, seq);
#ifdef DEBUG
	shLog("    generate_name: name is %s", name);
#endif

	/* 
	 * increment the sequence number so the next image won't 
	 *   overwrite this one 
	 */
	incr_sequence();

	return(0);
}


	/***************************************************************
	 * PROCEDURE: save_images
	 *
	 * DESCRIPTION: Save the images currently in the memory
	 *              card to FITS files on disk.  We will
	 *              call an external program to do the dirty
	 *              work -- all we have to do is give it
	 *              the proper parameters.
	 *
	 * RETURNS:
	 *    0         if all goes well
	 *    1         if there's a problem
	 */

	/* this character in an image file name describes the camera */
#define CAMERA_CHAR_INDEX  7

int
save_images
	(
	void
	)
{
	char template[LINELEN+1];
	char filename1[LINELEN+1];
	char filename2[LINELEN+1];
	char system_buffer[LINELEN+1];
	int retval;


	/* get a proper name for the images */
	if (generate_name(template) != 0) {
		shError("save_images: generate_name fails");
		return(1);
	}

	/*
	 * see TASS E-mail message 2001-06/msg00134.html for a 
	 * description of the filename format, and the meaning
	 * of the accepted codes for cameras
	 */
	/* this should be the I-band camera */
	strncpy(filename1, template, LINELEN);
	filename1[CAMERA_CHAR_INDEX] = 'F';
	/* this should be the V-band camera */
	strncpy(filename2, template, LINELEN);
	filename2[CAMERA_CHAR_INDEX] = 'E';

	/*
	 * prepare to make a system call to execute the external
	 * program which will read out the memory card and save
	 * images to disk
	 */
	sprintf(system_buffer, "%s/%s %d %d %s/%s %s/%s",
			CCD_PROGPATH, CCD_PROGNAME, CCD_NROW, CCD_NCOL,
			CCD_SAVEPATH, filename1, 
			CCD_SAVEPATH, filename2);

#ifdef DEBUG
	printf("about to system: ..%s..\n", system_buffer);
#endif
	retval = system(system_buffer);
	if (retval != 0) {
		shError("save_images: system returns %d", retval);
	}

	return(0);
}


	/***************************************************************
	 * PROCEDURE: set_exp_clock
	 *
	 * DESCRIPTION: Note the current system time and 
	 *              save for later reference.  We will use 
	 *              this when generating FITS header values.
	 *
	 *              This routine should be called just after
	 *              a clearing scan, just before the shutters
	 *              are opened.
	 *
	 * RETURNS:
	 *    nothing
	 */

void
set_exp_clock
	(
	void
	)
{
	exp_clock = time((time_t *)0);
}


	/***************************************************************
	 * PROCEDURE: get_exp_clock
	 *
	 * DESCRIPTION: Get the system time of the most recent
	 *              exposure (set by set_exp_clock).  We can
	 *              convert this time at our leisure to the
	 *              appropriate FITS keyword values.
	 *
	 * RETURNS:
	 *    saved system time (time_t)
	 */

static time_t
get_exp_clock
	(
	void
	)
{
	return(exp_clock);
}


	/***************************************************************
	 * PROCEDURE: set_exp_length
	 *
	 * DESCRIPTION: Save an exposure time for later reference.
	 *              We will use this when generating FITS header values.
	 *
	 *              This routine should be called every time
	 *              the user requests an exposure.
	 *
	 * RETURNS:
	 *    nothing
	 */

void
set_exp_length
	(
	int exptime          /* I: the exposure time in seconds */
	)
{
	exp_length = exptime;
}


	/***************************************************************
	 * PROCEDURE: get_exp_length
	 *
	 * DESCRIPTION: Get the exposure time of the most recent
	 *              exposure.
	 *
	 * RETURNS:
	 *    saved exposure time (seconds)
	 */

static int
get_exp_length
	(
	void
	)
{
	return(exp_length);
}


	/***************************************************************
	 * PROCEDURE: set_exp_type
	 *
	 * DESCRIPTION: Save an exposure type for later reference.
	 *              We will use this when generating FITS header values.
	 *              Typical types are 
	 *                   "dark", "flat", "object"
	 *
	 *              This routine should be called every time
	 *              the user requests an exposure.
	 *
	 * RETURNS:
	 *    nothing
	 */

void
set_exp_type
	(
	char *type          /* I: the exposure type */
	)
{
	strncpy(exp_type, type, LINELEN);
	exp_type[LINELEN] = '\0';
}


	/***************************************************************
	 * PROCEDURE: get_exp_type
	 *
	 * DESCRIPTION: Get the exposure type of the most recent
	 *              exposure.
	 *
	 * RETURNS:
	 *    saved exposure type
	 */

static char *
get_exp_type
	(
	void
	)
{
	return(exp_type);
}



	/****************************************************************
	 * PROCEDURE: set_hdr_keywords
	 *
	 * DESCRIPTION: Modify lines in the given "template" file,
	 *              which will be used by the "downfits" program
	 *              to write a FITS header.
	 *
	 *              We modify the file by replacing some of the
	 *              lines, but leaving others as-is.  The lines
	 *              we modify are
	 *
	 *                  DATE-OBS
	 *                  UT
	 *                  EXPTIME
	 *
	 *              Our method is:
	 *                  read in old file
	 *                  replace lines as necessary
	 *                  write modified version as temp file
	 *                  unlink old file
	 *                  rename temp file to old file name
	 *
	 * RETURNS:
	 *    0          if all goes well
	 *    1          if there's a problem
	 */

#define KEYWORD_LENGTH  8

static int
set_hdr_keywords
	(
	char *filename,          /* I: name of template file */
	char *hdr_dateobs,       /* I: string with DATE-OBS information */
	char *hdr_ut,            /* I: string with UT information */
	char *hdr_exptime,       /* I: string with EXPTIME information */
	char *hdr_imagetyp       /* I: string with IMAGETYP information */
	)
{
	char line[LINELEN+1];
	char tempfile[LINELEN+1];
	FILE *old_fp, *temp_fp;

	if ((old_fp = fopen(filename, "r")) == NULL) {
		shError("set_hdr_keywords: can't open file %s", filename);
		return(1);
	}
	strcpy(tempfile, "QQset_hdr.tmp");
	if ((temp_fp = fopen(tempfile, "w")) == NULL) {
		shError("set_hdr_keywords: can't open temp file %s", tempfile);
		return(1);
	}

	/* 
	 * read from the old, and write to the tempfile.
	 *   For each line we read, we write either the original,
	 *   or the corresponding supplied argument in its place.
	 */
	while (fgets(line, LINELEN, old_fp) != NULL) {
		if (strncmp(line, hdr_dateobs, KEYWORD_LENGTH) == 0) {
			fprintf(temp_fp, "%s\n", hdr_dateobs);
		}
		else if (strncmp(line, hdr_ut, KEYWORD_LENGTH) == 0) {
			fprintf(temp_fp, "%s\n", hdr_ut);
		}
		else if (strncmp(line, hdr_exptime, KEYWORD_LENGTH) == 0) {
			fprintf(temp_fp, "%s\n", hdr_exptime);
		}
		else if (strncmp(line, hdr_imagetyp, KEYWORD_LENGTH) == 0) {
			fprintf(temp_fp, "%s\n", hdr_imagetyp);
		}
		else {
			fprintf(temp_fp, "%s", line);
		}
	}
	fclose(temp_fp);
	fclose(old_fp);

	/* now, delete the old file and replace it with the temp file */
	if (unlink(filename) != 0) {
		shError("set_hdr_keywords: unlink fails on old %s", filename);
		return(1);
	}
	if (link(tempfile, filename) != 0) {
		shError("set_hdr_keywords: link fails to copy %s", filename);
		return(1);
	}
	if (unlink(tempfile) != 0) {
		shError("set_hdr_keywords: unlink fails on temp %s", tempfile);
		return(1);
	}

	return(0);
}




