
	/**************************************************************
	 * Functions to open, close, read, and write from/to a serial port.
	 * 
	 * MWR 7/29/2001
	 */

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>  /* String function definitions */
#include <fcntl.h>   /* File control definitions */
#include <errno.h>   /* Error number definitions */
#include <termios.h> /* POSIX terminal control definitions */
#include <sys/ioctl.h>
#include <sys/time.h>     /* for debugging time stuff */
#include "tait.h"


	/* global variables */
int serial_fd = -1;         /* file descriptor for the serial port */



	/*******************************************************************
	 * PROCEDURE: init_serial
	 *	
	 * DESCRIPTION: Set up the serial port for input and output.
	 *              We control the Mark IV by sending command
	 *              strings of 8 bytes to the port, and reading
	 *              strings of 8 bytes in return.
	 *
	 *              We set the global variable "serial_fd" to the
	 *              file descriptor we open here.  Routines which
	 *              talk to the serial port will use it.
	 *
	 * RETURNS:
	 *     0           if all goes well
	 *     1           if an error occurs
	 */

int
init_serial
	(
	void
	)
{
	struct termios options;


	serial_fd = open(SERIAL_DEV, O_RDWR | O_NOCTTY | O_NDELAY);
	if (serial_fd == -1) {
		/* Could not open the port. */
		perror("open_port: Unable to open serial port ");
	}
	else {
		fcntl(serial_fd, F_SETFL, 0);
	}

	/* get the current options for the port */
	tcgetattr(serial_fd, &options);

	/* Set the baud rates to 2400 ...  */
	cfsetispeed(&options, B2400);
	cfsetospeed(&options, B2400);

	/* choose raw input mode */
        options.c_lflag &= ~(ICANON | ECHO | ECHOE | ISIG);

	/* choose raw output mode */
	options.c_oflag &= ~OPOST;

	/* set 8N1 communications */
	options.c_cflag &= ~PARENB;
	options.c_cflag &= ~CSTOPB;
	options.c_cflag &= ~CSIZE;
	options.c_cflag |= CS8;

	/* and now we set the options */
	tcsetattr(serial_fd, TCSANOW, &options);


	return(0);
}

	

	/*******************************************************************
	 * PROCEDURE: close_serial
	 *	
	 * DESCRIPTION: Close the serial port.
	 *
	 * RETURNS:
	 *     0           if all goes well
	 *     1           if an error occurs
	 */

int
close_serial
	(
	void
	)
{

	close(serial_fd);

	return(0);
}

	

	/*******************************************************************
	 * PROCEDURE: send_serial
	 * 
	 * DESCRIPTION: Send an 8-byte message to the serial port,
	 *              and wait for a reply.  Place the reply
	 *              into the second argument.
	 *
	 *              As of 8/10/2001, the behavior of the STAMP seems
	 *              strange: after we send it a message, it almost
	 *              immediately (after 0.04 sec) sends us an exact
	 *              duplicate of the message; then, after a longer
	 *              pause (about 0.40 sec), it sends a proper reply.
	 *              I don't understand why this happens, but I'll
	 *              try to account for it in this routine.
	 *
	 * 
	 * RETURNS:
	 *      0        if all goes well
	 *      1        if an error occurs 
	 */


int
send_serial
	(
	char *message,         /* I: outgoing message: will send 8 bytes */
	                       /*        padding if necessary */
	char *reply            /* O: incoming reply: expect exactly 8 bytes, */
	                       /*        and will add NULL at 9'th char */
	                       /*        for convenience */
	)
{
	char message_copy[MSG_LEN + 3];
	char reply_copy[MSG_LEN + 3];
	int i;
	int nbyte;
	int done; 
	int first_char, last_char;
	int micro_sleep_time;     	

	/* 
	 * we must sleep this many microseconds each time we call
	 * this function -- otherwise, we aren't understood by
	 * the STAMP.
	 */
	micro_sleep_time = MICRO_SLEEP_TIME;
	usleep(micro_sleep_time);


	for (i = 0; i < MSG_LEN; i++) {
		message_copy[i] = message[i];
	}
	message_copy[MSG_LEN] = '\0';

	/* 
	 * enter loop: 
	 *      send message
	 *      wait for quick reply
	 *        verify that quick reply is duplicate of message
	 *        discard quick reply
	 *      wait for true reply
	 *      check to see if the 0'th and 7'th chars in reply
	 *           match the values in message
	 *      if yes, done
	 *      if no, go back to top of loop and re-send message
	 */
	done = 0;
	while (!done) {

		/* 
		 * we must sleep this many microseconds each time we call
		 * this function -- otherwise, we aren't understood by
		 * the STAMP.
		 */
		micro_sleep_time = MICRO_SLEEP_TIME;
		usleep(micro_sleep_time);
	
		/* must flush all serial I/O each time */
		tcflush(serial_fd, TCIOFLUSH);

		/* 
		 * pick two random characters between '0' and '9'
		 * and set the 0'th and 7'th chars of message_copy
		 * to those values.
		 */
		first_char = '0' + (int) 9*(random()/(double) RAND_MAX);
		last_char  = '0' + (int) 9*(random()/(double) RAND_MAX);
		printf("  first char %d = ..%c..   last %d = ..%c..\n",
				first_char, (char) first_char,
				last_char, (char) last_char);
		message_copy[0] = (char) first_char;
		message_copy[7] = (char) last_char;

#ifdef DEBUG
		printf("send_serial: about to send ..%s..\n", message_copy);
#endif
		nbyte = write(serial_fd, message_copy, MSG_LEN);
		if (nbyte != MSG_LEN) {
			fprintf(stderr, "send_serial: write failed");
			return(1);
		}
	
		/* prepare a place for the reply message */
		for (i = 0; i < MSG_LEN; i++) {
			reply_copy[i] = 'x';
		}
		reply_copy[MSG_LEN] = '\0';

		/* wait for the quick and spurious reply -- discard it */
		fcntl(serial_fd, F_SETFL, 0);
		nbyte = read(serial_fd, reply_copy, MSG_LEN);
		if (nbyte != 8) {
			shError("send_serial: quick read failed");
			continue;
		}
		if (strncmp(reply_copy, message_copy, MSG_LEN) != 0) {
			shError("send_serial: quick read not duplicate of message");
			continue;
		}
			
		/* now wait for the TRUE reply */
		fcntl(serial_fd, F_SETFL, 0);
		nbyte = read(serial_fd, reply_copy, MSG_LEN);
		if (nbyte != 8) {
			shError("send_serial: quick read failed");
			continue;
		}
#ifdef DEBUG
		printf("send_serial: reply is ..%s..\n", reply_copy);
#endif
	
/******************************************************************
{ int kk;
strncpy(reply, "xxxxxxxx", 8);
reply[8] = '\0';
for (kk = 0; kk < 3; kk++) {
		fcntl(serial_fd, F_SETFL, 0);
		for (i = 0; i < 10000000; i++) {
			if (i % 10000 == 0) {
				putchar('.');
			}
			ioctl(serial_fd, FIONREAD, &nbyte);
			if (nbyte >= MSG_LEN) {
				break;
			}
		}
		putchar('\n');
#if 0
		usleep(micro_sleep_time);
#endif
		nbyte = read(serial_fd, reply_copy, MSG_LEN);
		if (nbyte != 8) {
			fprintf(stderr, "send_serial: read failed");
			return(1);
		}
#ifdef DEBUG
		printf("send_serial: reply is ..%s..\n", reply_copy);
#endif
}
}
************************************************************/
	
		/* 
		 * check to make sure that the first and last chars
		 * of the reply are the same as the random values
		 * we chose at top of loop.  If not, we sleep and
		 * go back to top of loop.
		 */
		if ((reply_copy[0] != first_char) || 
		    (reply_copy[7] != last_char)) {
#ifdef DEBUG		
			printf("send_serial: first/last mismatch \n");
#endif
			micro_sleep_time = RETRY_SLEEP_TIME;
			usleep(micro_sleep_time);
		}
		else {
			/* okay, the reply has same first/last chars */
			done = 1;
		}
	}


	/* copy the reply into the output argument */
	for (i = 0; i < MSG_LEN; i++) {
		reply[i] = reply_copy[i];
	}
	reply[MSG_LEN] = '\0';

	/* all done */
	return(0);
}

