
	/* we need time information in a lot of places */
#include <time.h>

#define DEBUG

	/* 
	 * observatory latitude and longitude, both in decimal degrees:
	 *   latitude is positive North of the equator
	 *   longitude is positive EAST of Greenwich (so USA is all negative)
	 */
#define LATITUDE 43.0758 
#define LONGITUDE -77.6058


	/* input lines must be shorter than this many characters */
#define LINELEN        512

	/* message to and from the Stamp are this many bytes */
#define MSG_LEN  8

	/* 
	 * we must sleep for this many microseconds before every
	 * command we send to the STAMP over the serial port;
	 * otherwise, the STAMP doesn't receive the command properly
	 */
#define MICRO_SLEEP_TIME   250000
	/*
	 * If the stamp doesn't return a valid reply for a message,
	 * we sleep this many microseconds and try again.
	 */
#define RETRY_SLEEP_TIME  1000000

	/* name of the serial port connected to the Mark IV */
#define SERIAL_DEV     "/dev/ttyS0"

	/* 
	 * location of the base of the memory card in address space.
	 * We communicate with it by setting and reading bytes
	 * at the base location, and one byte higher 
	 */
#define MEMCARD_BASE   0x300

	/* 
	 * we place diagnostic messages (if any desired) into this file,
	 * to which new material is always appended.
	 */
#define LOGFILE "./tait.log"


	/*
	 * The "parse_and_execute" function returns one of these codes.
	 * PARSE_QUIT indicates that the program should terminate.
	 */
#define PARSE_OK      0
#define PARSE_ERROR   1
#define PARSE_QUIT    2


	/*
	 * we need this stuff to read out the CCDs properly.
	 */
#define CCD_NROW      2037
#define CCD_NCOL      2043
#define CCD_PROGPATH  "../downfits"
#define CCD_PROGNAME  "downfits"
#define CCD_SAVEPATH  "."
#define CCD_PLATEI    "./plate_1"
#define CCD_PLATEV    "./plate_2"
	/* 
	 * number of seconds which elapse between opening and closing
	 *   the V and I shutters when one requests a dual exposure.
	 *   In other words, if one requests a 0-second light exposure,
	 *   one actually gets an exposure this long.
	 */
#define EXPOSURE_BUILTIN          4.35
	/* 
	 * number of seconds which elapse between requesting a dual
	 *   exposure and the midpoint of opening I-band shutter.
	 */
#define EXPOSURE_IDELAY           0.90
	/* 
	 * number of seconds which elapse between requesting a dual
	 *   exposure and the midpoint of opening V-band shutter.
	 */
#define EXPOSURE_VDELAY           3.10

	/*
	 * These symbols are used to describe motor motions 
	 */
	/* motor actions */
typedef enum motor_action { 
  MOTOR_LIMIT_OUT    =   0,     /* move to limit, then out N steps */
  MOTOR_IN_DIRECTION =   1,     /* move in specified direction N steps */
  MOTOR_OFF          =   2      /* turn motor off */
} MOTOR_ACTION;
	/* motors we can control via "motor" function */
typedef enum motor_id {
  MOTOR_FOCUSV       =   0,     /* Focus V: limit sw pos 0, mask 0001 */
  MOTOR_DECDRIVE     =   1,     /* Dec drive:        pos 3, mask 0040 */
  MOTOR_DECCLAMP     =   2,     /* Dec clamp: no limit switch */
  MOTOR_FOCUSI       =   3      /* Focus I: limit sw pos 1, mask 0080 */
} MOTOR_ID;
	/* which direction to move motor? */
typedef enum motor_direction {
  MOTOR_TOWARD_LIMIT =   0,     /* move towards the limit switch */
  MOTOR_FROM_LIMIT   =   1      /* move away from the limit switch */
} MOTOR_DIRECTION;

	/* these describe servo motor stuff */
typedef enum servo_id {
  SERVO_VSHUTTER     =   0,     /* the servo controlling V camera's shutter */
  SERVO_ISHUTTER     =   1      /* the servo controlling I camera's shutter */
} SERVO_ID;
typedef enum servo_position {
  SERVO_CLOSESHUTTER  =  64,    /* move a shutter to the closed position */
  SERVO_OPENSHUTTER   = 255     /* move a shutter to the open position */
} SERVO_POSITION;


	/* global variables */
/* 
 * set this to 0 for no diagnostic messages, 
 *             1 for some messages
 *            >1 for lots of messages
 */
int debug;

/* 
 * if set to 1, then we are just simulating -- don't actually
 *   send commands to the Mark IV
 */
int g_fake_flag;

/* these are labels for the ADCs and DACs */
#define LABEL_LEN  50
char g_labels[100][LABEL_LEN+1];

/* command voltages for the DACs */
#define NUM_DACS   16
double g_voltages[NUM_DACS];

/* values for the ADCs */
#define NUM_ADCS   32
#define ADC_SIZE    4
double g_adc[NUM_ADCS][ADC_SIZE];

/* pulse control for the CCD electronics */
#define NUM_PULSE  24
#define PULSE_SIZE  2
char g_pulse[NUM_PULSE][PULSE_SIZE+1];


	/* declarations of global routines */
int send_message(char *message, char *reply, char *comment);

int get_next_line(char *filename, char *first_line);

void init_globals(void);
int init_serial(void);
int close_serial(void);
int send_serial(char *message, char *reply);

int load_dacs(void);
int load_one_dac(int dac_index, double voltage);
void show_dacs(void);
int read_adcs(void);
void show_adcs(void);

void *shMalloc(int nbytes);
void shFree(void *vptr);
void shError(char *format, ...);
void shFatal(char *format, ...);
void shLog(char *format, ...);

int read_next_command(char *input_file, char *command_line);
int parse_and_execute(char *cmdline);
void init_hash(void);

int motor(MOTOR_ACTION action, MOTOR_ID motor_id, 
		MOTOR_DIRECTION direction, int steps);

int dec_is_at_limit(void);
int ra_is_at_limit(void);
int foc_is_at_limit(MOTOR_ID id);

int servo(SERVO_ID servo_id, SERVO_POSITION position);

int init_memcard(void);
int scan_both(void);

void init_sequence(void);
int set_sequence(int sequence_number);
int get_sequence(void);
int generate_name(char *name);

void get_now(struct tm *tms);
void get_now_jd(double *jd);
void get_jd(int day, int month, int year, double ut, double *jd);
void get_lst(double jd, double *lst);
void get_ha(double ra, double *ha);

int save_images(void);
void set_exp_clock(void);
void set_exp_length(int exptime);
void set_exp_type(char *exptype);


   /*
    * This is a preprocessor macro, which acts like a function.  The user
    * calls it with one argument, a condition which should evaluate to
    * 0 or 1.  If it evaluates to 1, then nothing happens; but if it
    * evaluates to 0, then the program prints an error message, giving
    * location of the error, and halts execution with an error code.
    * 
    * Thus, one typically uses it to make a 'sanity check', such as
    * making sure that a pointer (which really, really should have
    * a valid value) isn't NULL:
    *
    *     fp = fopen("file", "r");
    *        ...
    *     shAssert(fp != NULL);
    *     fgets(line, LINELEN, fp);
    *        ...
    */
#include <stdlib.h>
#define shAssert(x) if((x)!=1){fprintf(stderr,"assertion fails in file %s, line %d\n",__FILE__,__LINE__);exit(1);}
 

