
	/* 
	 * this file contains routines for determining various astronomical 
	 * times and dates.
	 *
	 * It also holds the values of RA and Dec which were last requested
	 * by the user.  We keep those for future reference, so we know
	 * where the telescope ought to be pointing.
	 */

	/*
	 * Change "BASE" in since_1950 from 2433282.5 to 2433281.5 -- it
	 * represents the Julian Date at Jan 0, 1950, at 0 hours UT.
	 * Previous value gave Julian Dates which were always 1 day too
	 * large.
	 * MWR 7/6/2001
	 */

#include <stdio.h>
#include <time.h>
#include <math.h>
#include "tait.h"

	/* 
	 * These keep track of the last (RA, Dec) the user requested.
	 *   The variables can be read and set only by the routines
	 *             save_telescope_ra   retrieve_telescope_ra
	 *             save_telescope_dec  retrieve_telescope_dec
	 */
static double last_set_ra  = -999.0;
static double last_set_dec = -999.0;


static int since_1950(int day, int month, int year);
static int is_leap_year(int year);


	/******************************************************************
	 * PROCEDURE: get_now
	 *
	 * DESCRIPTION: get the current UT date and UT time.  The information
	 *              is placed into a "struct tm" structure, which the
	 *              calling function must provide.
	 *
	 * RETURNS:
	 *    nothing
	 */

void
get_now
	(
	struct tm *tms                /* O: we fill the fields in this struct */
	)
{
	struct tm *local_tms;
	time_t clock;

	/* first, we get the current time in seconds from Jan 1, 1970 */
	clock = time((time_t *) NULL);

	/* now, convert to UT date and time */
	local_tms = gmtime(&clock);
	tms->tm_sec = local_tms->tm_sec;
	tms->tm_min = local_tms->tm_min;
	tms->tm_hour = local_tms->tm_hour;
	tms->tm_mday = local_tms->tm_mday;
	tms->tm_mon = local_tms->tm_mon;
	tms->tm_year = local_tms->tm_year;
	tms->tm_wday = local_tms->tm_wday;
	tms->tm_yday = local_tms->tm_yday;
	tms->tm_isdst = local_tms->tm_isdst;
}

	/******************************************************************
	 * PROCEDURE: get_now_jd
	 *
	 * DESCRIPTION: get the current Julian Date.  The JD is placed
	 *              is the passed argument.
	 *
	 * RETURNS:
	 *    nothing
	 */

void
get_now_jd
	(
	double *jd                /* O: the current JD will be placed here */
	)
{
	int year, month, day;
	double ut;
	struct tm *local_tms;
	time_t clock;

	/* first, we get the current time in seconds from Jan 1, 1970 */
	clock = time((time_t *) NULL);

	/* now, convert to UT date and time */
	local_tms = gmtime(&clock);

	/* now, convert UT date and time to Julian Date */
	year = local_tms->tm_year + 1900;
	month = local_tms->tm_mon;
	day = local_tms->tm_mday;
	ut = local_tms->tm_hour + (local_tms->tm_min/60.0) + 
		             (local_tms->tm_sec/3600.0);
	get_jd(day, month, year, ut, jd);
}




	/******************************************************************
	 * PROCEDURE: get_jd
	 *
	 * DESCRIPTION: place the fractional Julian Date, which corresponds 
	 *              to the given day, month, year and UT time, 
	 *              into the passed "jd" argument.
	 *
	 * RETURNS:
	 *    0         if all goes well
	 *    1         if there's a problem
	 */

#define BASE     2433281.5         	/* JD of January 0.0 1950 */

void
get_jd(day, month, year, ut, jd)
int day, month, year;
double ut, *jd;
{
	double d;
	
	d = (double) since_1950(day, month, year);
	*jd = d + BASE + (ut/24.0);
}


	/********************************************************************
	 * PROCEDURE: since_1950
	 *
	 * DESCRIPTION: return the number of days between the given 
	 *              date and Jan 0, 1950 
	 *
	 * RETURNS:
	 *    number of days since Jan 0, 1950
	 */

static int 
since_1950
	(
	int day,             /* I: day within the month */
	int month,           /* I: month within year (0-11) */
	int year             /* I: year */
	)
{
	int i, sum;
	static int days[12] = { 31, 28, 31, 30, 31, 30, 
	                        31, 31, 30, 31, 30, 31 };

	sum = 0;
	for (i = 1950; i < year; i++) {
		sum += 365;
		/* check for leap year ... */
		if (is_leap_year(i)) {
			sum++;
		}
	}
	for (i = 1; i <= month; i++) {
		sum += days[i - 1];
		if ((i == 2) && (is_leap_year(year))) {
			sum++;
		}
	}
	sum += day;
	return(sum);
}


	/**********************************************************************
	 * PROCEDURE: is_leap_year
	 *
	 * DESCRIPTION: Given an integer year value (like 1954 or 2001), 
	 *              decide it it's a leap year.  
	 *
	 * RETURNS:
	 *    1            if it is a leap year
	 *    0            if not 
	 */

static int
is_leap_year
	(
	int year
	)
{
	if (year % 4 != 0) {
		return(0);
	}

	/* if we get here, we must have a year divisible by 4 */
	/* if it's not divisible by 100, then it's definitely a leap year */
	if (year % 100 != 0) {
		return(1);
	}

	/* 
	 * if we get here, we have a century year: 1900, 2000, 2100, etc. 
	 * If the century is divisible by 400, it's a leap year; 
	 * otherwise, not 
	 */
	if (year % 400 == 0) {
		return(1);
	} else {
		return(0);
	}

	/* we should never reach here */
	shFatal("is_leap_year: fell to bottom of function for year %d?!\n",
			year);
	exit(1);
}


/************************************************************************
 * PROCEDURE: get_lst
 *
 * DESCRIPTION: place the LST for the given Julian Date 
 *              into the passed variable.  Uses the procedure 	
 *              given by Duffett-Smith's "Practical Astronomy with 
 *              your Pocket Calculator", p. 17 and 20. 
 *              The observatory longitude must be converted from
 *              decimal degrees to HOURS WEST OF GREENWICH.
 *
 *              The LST is set in the range (0.0-23.99). 
 *
 * RETURNS:
 *    nothing
 */

void
get_lst(jd, lst)
double jd, *lst;
{
	double j, s, t, t0, u, longitude;

	j = floor(jd + 0.5) - 0.5;
	s = j - 2451545.0;
	t = s/36525.0;
	t0 = 6.697374558 + 2400.051336*t + 0.000025862*t*t;
	while (t0 > 24.0)
		t0 -= 24.0;
	while (t0 < 0.0)
		t0 += 24.0;
	u = (jd - j)*24.0;
	u *= 1.002737909;
	t0 += u;
	while (t0 > 24.0)
		t0 -= 24.0;
	while (t0 < 0.0)
		t0 += 24.0;

	/* 
	 * we need longitude in hours WEST of Greenwich -- so we need to 
	 *   convert from the #define'd value, which is degrees EAST of
	 *   Greenwich.
	 */
	longitude = LONGITUDE;
	longitude /= -15.0;

	t0 -= longitude;
	while (t0 > 24.0)
		t0 -= 24.0;
	while (t0 < 0.0)
		t0 += 24.0;

	*lst = t0;
}
	

	/*************************************************************************
	 * PROCEDURE: get_ha
	 *
	 * DESCRIPTION: Given an RA, use the current time to figure out
	 *              the Hour Angle (HA) of the position in the sky.
	 *              Express the HA in decimal degrees, within the
	 *              range -180 to +180 degrees.
	 *
	 *              If desired RA is West of meridian, HA > 0
	 *              If desired RA is East of meridian, HA < 0
	 *
	 * RETURNS:
	 *    nothing
	 */

void
get_ha
	(
	double ra,              /* I: desired Right Ascension, decimal degrees */
	double *ha              /* O: place Hour Angle here, decimal degrees */
	)
{
	double jd, lst;
	double local_ha;

	/* 
	 *   Calculate the Hour Angle for the given RA --
	 *   we need to get the LST to do this, and then calculate
	 *                HA = LST - RA
	 *
	 *   We want to end up with HA in decimal degrees, West of meridian
	 */
	get_now_jd(&jd);
	get_lst(jd, &lst);
	/* get_lst yields an LST in decimal hours; we must convert to degrees */
	lst *= 15.0;
	local_ha = lst - ra;

	/* make sure that the HA lies between -180 and 180 degrees */
	while (local_ha > 180) {
		local_ha -= 180.0;
	}
	while (local_ha < -180) {
		local_ha += 180.0;
	}

	*ha = local_ha;
}


	/***********************************************************************
	 * PROCEDURE: set_telescope_ra
	 *
	 * DESCRIPTION: Hold the given RA (decimal degrees) in a safe place,
	 *              so that we can ask for it later.  
	 *
	 *              The idea is that when the user requests that the
	 *              telescope point to a new spot in the sky, we save
	 *              the RA and Dec with these functions.  The telescope
	 *              will usually track the given spot for a while.
	 *              At some later time, we may want to know if the
	 *              current position is close to or exceeding the 
	 *              East/West limits -- we can then ask for the 
	 *              saved RA value to check.
	 *
	 * RETURNS:
	 *   nothing
	 */

void
save_telescope_ra
	(
	double ra              /* I: Right Ascension, in decimal degrees */
	)
{


}
